# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from typing import List, Optional


class DeleteParameters(dict):
    """
    Contains all parameters required to support DELETEs
    equality_column_names: List of column names that would be used to determine row equality for equality deletes.  Relevant only to equality deletes
    """

    @staticmethod
    def of(
        equality_column_names: Optional[List[str]] = None,
    ) -> DeleteParameters:
        delete_parameters = DeleteParameters()
        if equality_column_names is not None:
            delete_parameters["equality_column_names"] = equality_column_names
        return delete_parameters

    @property
    def equality_column_names(self) -> Optional[List[str]]:
        return self.get("equality_column_names")

    @staticmethod
    def merge_delete_parameters(
        delete_parameters: List[DeleteParameters],
    ) -> Optional[DeleteParameters]:
        if len(delete_parameters) < 2:
            return delete_parameters
        equality_column_names = delete_parameters[0].equality_column_names
        assert all(
            delete_prev.equality_column_names == delete_curr.equality_column_names
            for delete_prev, delete_curr in zip(
                delete_parameters, delete_parameters[1:]
            )
        ), "We cannot merge two delete parameters if their equality column names are different."
        merge_delete_parameters = DeleteParameters.of(equality_column_names)
        return merge_delete_parameters



from typing import List, Optional, Union
import pyarrow as pa
from deltacat.storage import Delta, Partition, PartitionLocator, DeltaLocator
import deltacat.tests.local_deltacat_storage as ds
from deltacat.types.media import StorageType, ContentType


def create_delta_from_csv_file(
    namespace: str,
    file_paths: List[str],
    table_name: Optional[str] = None,
    table_version: int = 1,
    content_type: ContentType = ContentType.PARQUET,
    *args,
    **kwargs,
) -> Delta:
    staged_partition = stage_partition_from_file_paths(
        namespace,
        file_paths,
        *args,
        table_name=table_name,
        table_version=table_version,
        **kwargs,
    )
    committed_delta = commit_delta_to_staged_partition(
        staged_partition, file_paths, content_type=content_type, *args, **kwargs
    )
    return committed_delta


def stage_partition_from_file_paths(
    namespace: str,
    file_paths: List[str],
    table_name: Optional[str] = None,
    table_version: int = 1,
    *args,
    **kwargs,
) -> Partition:
    ds.create_namespace(namespace, {}, **kwargs)
    if table_name is None:
        table_name = "-".join(file_paths).replace("/", "_")
    ds.create_table_version(namespace, table_name, str(table_version), **kwargs)
    stream = ds.get_stream(namespace, table_name, str(table_version), **kwargs)
    staged_partition = ds.stage_partition(stream, [], **kwargs)
    return staged_partition


def commit_delta_to_staged_partition(
    staged_partition,
    file_paths: List[str],
    content_type: ContentType = ContentType.PARQUET,
    *args,
    **kwargs,
) -> Delta:
    committed_delta = commit_delta_to_partition(
        staged_partition,
        *args,
        file_paths=file_paths,
        content_type=content_type,
        **kwargs,
    )
    ds.commit_partition(staged_partition, **kwargs)
    return committed_delta


def download_delta(delta_like: Union[Delta, DeltaLocator], *args, **kwargs) -> Delta:
    return pa.concat_tables(
        ds.download_delta(
            delta_like,
            storage_type=StorageType.LOCAL,
            *args,
            **kwargs,
        )
    )


def commit_delta_to_partition(
    partition: Union[Partition, PartitionLocator],
    file_paths: List[str],
    content_type: ContentType = ContentType.PARQUET,
    *args,
    **kwargs,
) -> Delta:
    tables = []

    if isinstance(partition, PartitionLocator):
        partition = ds.get_partition(
            partition.stream_locator, partition.partition_values, *args, **kwargs
        )

    for file_path in file_paths:
        table = pa.csv.read_csv(file_path)
        tables.append(table)

    table = pa.concat_tables(tables)
    staged_delta = ds.stage_delta(table, partition, content_type=content_type, **kwargs)

    return ds.commit_delta(staged_delta, **kwargs)



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

import itertools
import logging
from typing import Any, Dict, List, Optional
from uuid import uuid4
from enum import Enum

from deltacat import logs

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


class EntryType(str, Enum):
    """
    Enum representing all possible content categories of an manifest entry file
    """

    DATA = "data"
    POSITIONAL_DELETE = "positional_delete"
    EQUALITY_DELETE = "equality_delete"

    @classmethod
    def get_default(cls):
        return EntryType.DATA

    @classmethod
    def list(cls):
        return [c.value for c in EntryType]


class EntryFileParams(dict):
    """
    Represents parameters relevant to the underlying contents of manifest entry. Contains all parameters required to support DELETEs
    equality_column_names: List of column names that would be used to determine row equality for equality deletes.  Relevant only to equality deletes
    position: Ordinal position of a deleted row in the target data file identified by uri, starting at 0. Relevant only to positional deletes
    """

    @staticmethod
    def of(
        equality_column_names: Optional[List[str]] = None,
        position: Optional[int] = None,
    ) -> EntryFileParams:
        entry_file_params = EntryFileParams()
        if equality_column_names is not None:
            entry_file_params["equality_column_names"] = equality_column_names
        if position is not None:
            entry_file_params["position"] = position
        return entry_file_params

    @property
    def equality_column_names(self) -> Optional[List[str]]:
        return self.get("equality_column_names")

    @property
    def url(self) -> Optional[str]:
        return self.get("url")

    @property
    def position(self) -> Optional[int]:
        return self.get("position")


class Manifest(dict):
    @staticmethod
    def _build_manifest(
        meta: Optional[ManifestMeta],
        entries: Optional[ManifestEntryList],
        author: Optional[ManifestAuthor] = None,
        uuid: str = None,
        entry_type: Optional[EntryType] = None,
    ) -> Manifest:
        if not uuid:
            uuid = str(uuid4())
        manifest = Manifest()
        manifest["id"] = uuid
        if meta is not None:
            manifest["meta"] = meta
        if entries is not None:
            manifest["entries"] = entries
        if author is not None:
            manifest["author"] = author
        if entry_type is not None:
            manifest["entry_type"] = entry_type.value
        return manifest

    @staticmethod
    def of(
        entries: ManifestEntryList,
        author: Optional[ManifestAuthor] = None,
        uuid: str = None,
        entry_type: Optional[EntryType] = None,
    ) -> Manifest:
        if not uuid:
            uuid = str(uuid4())
        total_record_count = 0
        total_content_length = 0
        total_source_content_length = 0
        content_type = None
        content_encoding = None
        partition_values_set = set()
        partition_values = None
        if entries:
            content_type = entries[0].meta.content_type
            content_encoding = entries[0].meta.content_encoding
            for entry in entries:
                meta = entry.meta
                if meta.content_type != content_type:
                    content_type = None
                if meta.content_encoding != content_encoding:
                    content_encoding = None
                entry_content_type = meta.content_type
                if entry_content_type != content_type:
                    msg = (
                        f"Expected all manifest entries to have content "
                        f"type '{content_type}' but found "
                        f"'{entry_content_type}'"
                    )
                    raise ValueError(msg)
                entry_content_encoding = meta["content_encoding"]
                if entry_content_encoding != content_encoding:
                    msg = (
                        f"Expected all manifest entries to have content "
                        f"encoding '{content_encoding}' but found "
                        f"'{entry_content_encoding}'"
                    )
                    raise ValueError(msg)
                total_record_count += meta.record_count or 0
                total_content_length += meta.content_length or 0
                total_source_content_length += meta.source_content_length or 0
                if len(partition_values_set) <= 1:
                    partition_values_set.add(entry.meta.partition_values)

        if len(partition_values_set) == 1:
            partition_values = partition_values_set.pop()

        meta = ManifestMeta.of(
            total_record_count,
            total_content_length,
            content_type,
            content_encoding,
            total_source_content_length,
            entry_type=entry_type,
            partition_values=partition_values,
        )
        manifest = Manifest._build_manifest(meta, entries, author, uuid, entry_type)
        return manifest

    @staticmethod
    def merge_manifests(
        manifests: List[Manifest], author: Optional[ManifestAuthor] = None
    ) -> Manifest:
        all_entries = ManifestEntryList(
            itertools.chain(*[m.entries for m in manifests])
        )
        merged_manifest = Manifest.of(all_entries, author)
        return merged_manifest

    @property
    def meta(self) -> Optional[ManifestMeta]:
        val: Dict[str, Any] = self.get("meta")
        if val is not None and not isinstance(val, ManifestMeta):
            self["meta"] = val = ManifestMeta(val)
        return val

    @property
    def entries(self) -> Optional[ManifestEntryList]:
        val: List[ManifestEntry] = self.get("entries")
        if val is not None and not isinstance(val, ManifestEntryList):
            self["entries"] = val = ManifestEntryList.of(val)
        return val

    @property
    def id(self) -> str:
        return self["id"]

    @property
    def author(self) -> Optional[ManifestAuthor]:
        val: Dict[str, Any] = self.get("author")
        if val is not None and not isinstance(val, ManifestAuthor):
            self["author"] = val = ManifestAuthor(val)
        return val


class ManifestMeta(dict):
    @staticmethod
    def of(
        record_count: Optional[int],
        content_length: Optional[int],
        content_type: Optional[str],
        content_encoding: Optional[str],
        source_content_length: Optional[int] = None,
        credentials: Optional[Dict[str, str]] = None,
        content_type_parameters: Optional[List[Dict[str, str]]] = None,
        entry_type: Optional[EntryType] = None,
        partition_values: Optional[List[str]] = None,
    ) -> ManifestMeta:
        manifest_meta = ManifestMeta()
        if record_count is not None:
            manifest_meta["record_count"] = record_count
        if content_length is not None:
            manifest_meta["content_length"] = content_length
        if source_content_length is not None:
            manifest_meta["source_content_length"] = source_content_length
        if content_type is not None:
            manifest_meta["content_type"] = content_type
        if content_type_parameters is not None:
            manifest_meta["content_type_parameters"] = content_type_parameters
        if content_encoding is not None:
            manifest_meta["content_encoding"] = content_encoding
        if credentials is not None:
            manifest_meta["credentials"] = credentials
        if entry_type is not None:
            manifest_meta["entry_type"] = entry_type.value
        if partition_values is not None:
            manifest_meta["partition_values"] = partition_values
        return manifest_meta

    @property
    def record_count(self) -> Optional[int]:
        return self.get("record_count")

    @property
    def content_length(self) -> Optional[int]:
        return self.get("content_length")

    @property
    def content_type(self) -> Optional[str]:
        return self.get("content_type")

    @property
    def content_encoding(self) -> Optional[str]:
        return self.get("content_encoding")

    @property
    def source_content_length(self) -> Optional[int]:
        return self.get("source_content_length")

    @property
    def content_type_parameters(self) -> Optional[List[Dict[str, str]]]:
        return self.get("content_type_parameters")

    @content_type_parameters.setter
    def content_type_parameters(self, params: List[Dict[str, str]]) -> None:
        self["content_type_parameters"] = params

    @property
    def credentials(self) -> Optional[Dict[str, str]]:
        return self.get("credentials")

    @property
    def entry_type(self) -> Optional[EntryType]:
        val = self.get("entry_type")
        if val is not None:
            return EntryType(self["entry_type"])
        return val

    @property
    def partition_values(self) -> Optional[List[str]]:
        return self.get("partition_values")


class ManifestAuthor(dict):
    @staticmethod
    def of(name: Optional[str], version: Optional[str]) -> ManifestAuthor:
        manifest_author = ManifestAuthor()
        if name is not None:
            manifest_author["name"] = name
        if version is not None:
            manifest_author["version"] = version
        return manifest_author

    @property
    def name(self) -> Optional[str]:
        return self.get("name")

    @property
    def version(self) -> Optional[str]:
        return self.get("version")


class ManifestEntry(dict):
    @staticmethod
    def of(
        url: Optional[str],
        meta: Optional[ManifestMeta],
        mandatory: bool = True,
        uri: Optional[str] = None,
        uuid: Optional[str] = None,
        entry_type: Optional[EntryType] = None,
        entry_file_params: Optional[EntryFileParams] = None,
    ) -> ManifestEntry:
        manifest_entry = ManifestEntry()
        if not (uri or url):
            raise ValueError("No URI or URL specified for manifest entry contents.")
        if (uri and url) and (uri != url):
            raise ValueError(f"Manifest entry URI ({uri}) != URL ({url})")
        if url:
            manifest_entry["url"] = manifest_entry["uri"] = url
        elif uri:
            manifest_entry["url"] = manifest_entry["uri"] = uri
        if meta is not None:
            manifest_entry["meta"] = meta
        if mandatory is not None:
            manifest_entry["mandatory"] = mandatory
        if uuid is not None:
            manifest_entry["id"] = uuid
        if entry_type is not None:
            manifest_entry["entry_type"] = entry_type.value
        if entry_file_params is not None:
            if entry_file_params.get("url") != manifest_entry.get("url"):
                msg = (
                    f"Expected manifest entry url: {manifest_entry.url}"
                    f" and entry_file_params: '{entry_file_params.url}' to match"
                )
                raise ValueError(msg)
            manifest_entry["entry_file_params"] = entry_file_params
        return manifest_entry

    @staticmethod
    def from_s3_obj_url(
        url: str,
        record_count: int,
        source_content_length: Optional[int] = None,
        **s3_client_kwargs,
    ) -> ManifestEntry:
        from deltacat.aws import s3u as s3_utils

        s3_obj = s3_utils.get_object_at_url(url, **s3_client_kwargs)
        logger.debug(f"Building manifest entry from {url}: {s3_obj}")
        manifest_entry_meta = ManifestMeta.of(
            record_count,
            s3_obj["ContentLength"],
            s3_obj["ContentType"],
            s3_obj["ContentEncoding"],
            source_content_length,
        )
        manifest_entry = ManifestEntry.of(url, manifest_entry_meta)
        return manifest_entry

    @property
    def uri(self) -> Optional[str]:
        return self.get("uri")

    @property
    def url(self) -> Optional[str]:
        return self.get("url")

    @property
    def meta(self) -> Optional[ManifestMeta]:
        val: Dict[str, Any] = self.get("meta")
        if val is not None and not isinstance(val, ManifestMeta):
            self["meta"] = val = ManifestMeta(val)
        return val

    @property
    def mandatory(self) -> bool:
        return self["mandatory"]

    @property
    def id(self) -> Optional[str]:
        return self.get("id")

    @property
    def entry_type(self) -> Optional[EntryType]:
        val = self.get("entry_type")
        if val is not None:
            return EntryType(self["entry_type"])
        return val

    @property
    def entry_file_params(self) -> Optional[EntryFileParams]:
        val: Dict[str, Any] = self.get("entry_file_params")
        if val is not None and not isinstance(val, EntryFileParams):
            self["entry_file_params"] = val = EntryFileParams(val)
        return val


class ManifestEntryList(List[ManifestEntry]):
    @staticmethod
    def of(entries: List[ManifestEntry]) -> ManifestEntryList:
        manifest_entries = ManifestEntryList()
        for entry in entries:
            if entry is not None and not isinstance(entry, ManifestEntry):
                entry = ManifestEntry(entry)
            manifest_entries.append(entry)
        return manifest_entries

    def __getitem__(self, item):
        val = super().__getitem__(item)
        if val is not None and not isinstance(val, ManifestEntry):
            self[item] = val = ManifestEntry(val)
        return val



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations
from enum import Enum
from typing import Any, Dict, List, Optional, Set
import datetime as dt
from boto3.resources.base import ServiceResource
from datetime import timezone

from deltacat.tests.compute.test_util_constant import (
    TEST_S3_RCF_BUCKET_NAME,
)
from deltacat.tests.compute.test_util_constant import (
    BASE_TEST_SOURCE_NAMESPACE,
    BASE_TEST_SOURCE_TABLE_NAME,
    BASE_TEST_SOURCE_TABLE_VERSION,
    BASE_TEST_DESTINATION_NAMESPACE,
    BASE_TEST_DESTINATION_TABLE_NAME,
    BASE_TEST_DESTINATION_TABLE_VERSION,
    REBASING_NAMESPACE,
    REBASING_TABLE_NAME,
    REBASING_TABLE_VERSION,
)
from deltacat.compute.compactor import (
    RoundCompletionInfo,
)
from deltacat.compute.compactor.model.compaction_session_audit_info import (
    CompactionSessionAuditInfo,
)

from deltacat.storage.model.partition import PartitionLocator
from deltacat.storage.model.stream import StreamLocator
from deltacat.storage.model.table_version import TableVersionLocator
from deltacat.storage.model.table import TableLocator
from deltacat.storage.model.namespace import NamespaceLocator
from deltacat.compute.compactor.model.compactor_version import CompactorVersion


class PartitionKeyType(str, Enum):
    INT = "int"
    STRING = "string"
    TIMESTAMP = "timestamp"


class PartitionKey(dict):
    @staticmethod
    def of(key_name: str, key_type: PartitionKeyType) -> PartitionKey:
        return PartitionKey({"keyName": key_name, "keyType": key_type.value})

    @property
    def key_name(self) -> str:
        return self["keyName"]

    @property
    def key_type(self) -> PartitionKeyType:
        key_type = self["keyType"]
        return None if key_type is None else PartitionKeyType(key_type)


"""
UTILS
"""


def get_test_partition_locator(partition_id):
    tv_locator = TableVersionLocator.of(
        TableLocator.of(NamespaceLocator.of("default"), "test_table"), "1"
    )
    stream_locator = StreamLocator.of(tv_locator, "test_stream_id", "local")
    partition_locator = PartitionLocator.of(
        stream_locator, partition_id=partition_id, partition_values=[]
    )

    return partition_locator


def _create_table(
    namespace: str,
    table_name: str,
    table_version: str,
    primary_keys: Set[str],
    sort_keys: Optional[List[Any]],
    partition_keys: Optional[List[PartitionKey]],
    ds_mock_kwargs: Optional[Dict[str, Any]],
):
    import deltacat.tests.local_deltacat_storage as ds
    from deltacat.types.media import ContentType

    ds.create_namespace(namespace, {}, **ds_mock_kwargs)
    ds.create_table_version(
        namespace,
        table_name,
        table_version,
        primary_key_column_names=list(primary_keys),
        sort_keys=sort_keys,
        partition_keys=partition_keys,
        supported_content_types=[ContentType.PARQUET],
        **ds_mock_kwargs,
    )
    return namespace, table_name, table_version


def create_src_table(
    primary_keys: Set[str],
    sort_keys: Optional[List[Any]],
    partition_keys: Optional[List[PartitionKey]],
    ds_mock_kwargs: Optional[Dict[str, Any]],
):
    source_namespace: str = BASE_TEST_SOURCE_NAMESPACE
    source_table_name: str = BASE_TEST_SOURCE_TABLE_NAME
    source_table_version: str = BASE_TEST_SOURCE_TABLE_VERSION
    return _create_table(
        source_namespace,
        source_table_name,
        source_table_version,
        primary_keys,
        sort_keys,
        partition_keys,
        ds_mock_kwargs,
    )


def create_destination_table(
    primary_keys: Set[str],
    sort_keys: Optional[List[Any]],
    partition_keys: Optional[List[PartitionKey]],
    ds_mock_kwargs: Optional[Dict[str, Any]],
):
    destination_namespace: str = BASE_TEST_DESTINATION_NAMESPACE
    destination_table_name: str = BASE_TEST_DESTINATION_TABLE_NAME
    destination_table_version: str = BASE_TEST_DESTINATION_TABLE_VERSION
    return _create_table(
        destination_namespace,
        destination_table_name,
        destination_table_version,
        primary_keys,
        sort_keys,
        partition_keys,
        ds_mock_kwargs,
    )


def create_rebase_table(
    primary_keys: Set[str],
    sort_keys: Optional[List[Any]],
    partition_keys: Optional[List[PartitionKey]],
    ds_mock_kwargs: Optional[Dict[str, Any]],
):
    rebasing_namespace = REBASING_NAMESPACE
    rebasing_table_name = REBASING_TABLE_NAME
    rebasing_table_version = REBASING_TABLE_VERSION
    return _create_table(
        rebasing_namespace,
        rebasing_table_name,
        rebasing_table_version,
        primary_keys,
        sort_keys,
        partition_keys,
        ds_mock_kwargs,
    )


def get_rcf(s3_resource, rcf_file_s3_uri: str) -> RoundCompletionInfo:
    from deltacat.tests.test_utils.utils import read_s3_contents

    _, rcf_object_key = rcf_file_s3_uri.strip("s3://").split("/", 1)
    rcf_file_output: Dict[str, Any] = read_s3_contents(
        s3_resource, TEST_S3_RCF_BUCKET_NAME, rcf_object_key
    )
    return RoundCompletionInfo(**rcf_file_output)


def get_compacted_delta_locator_from_rcf(
    s3_resource: ServiceResource, rcf_file_s3_uri: str
):
    from deltacat.storage import DeltaLocator

    round_completion_info: RoundCompletionInfo = get_rcf(s3_resource, rcf_file_s3_uri)

    compacted_delta_locator: DeltaLocator = (
        round_completion_info.compacted_delta_locator
    )
    return compacted_delta_locator


def offer_iso8601_timestamp_list(
    periods: int,
    unit_of_time: str,
    end_time=dt.datetime(2023, 5, 3, 10, 0, 0, 0, tzinfo=timezone.utc),
) -> List[str]:
    """
    Returns a list of ISO 8601 timestamps, each periods units of time before the start time.

    Args:
    periods: The number of timestamps to return.
    unit_of_time: The unit of time to use for the timestamps. Must be one of "seconds", "minutes", "hours", "days", or "weeks".
    end_time: The end time for the timestamps. Defaults to 2023-05-03T10:00:00Z.

    Returns:
    A list of ISO 8601 timestamps, each periods units of time before the start time.

    Raises:
    ValueError: If the unit_of_time argument is not one of "seconds", "minutes", "hours", "days", or "weeks".
    """
    import datetime as dt

    UTC_ISO_8601_FORMAT_WITHOUT_MILLIS = "%Y-%m-%dT%H:%M:%SZ"  # '2018-09-05T14:09:03Z'

    acceptable_units_of_time = ["seconds", "minutes", "hours", "days", "weeks"]
    if unit_of_time not in acceptable_units_of_time:
        raise ValueError(
            f"unit_of_time {unit_of_time} is not supported. Please use one of these time units: {acceptable_units_of_time}"
        )
    res = []
    for i in range(periods):
        kwarg = {unit_of_time: i}
        res.append(
            (end_time - dt.timedelta(**kwarg)).strftime(
                UTC_ISO_8601_FORMAT_WITHOUT_MILLIS
            )
        )
    return res


def assert_compaction_audit(
    compactor_version: CompactorVersion, compaction_audit: CompactionSessionAuditInfo
):
    if compactor_version == CompactorVersion.V2:
        audit_entries = [
            compaction_audit.deltacat_version,
            compaction_audit.ray_version,
            compaction_audit.audit_url,
            compaction_audit.hash_bucket_count,
            compaction_audit.compactor_version,
            compaction_audit.input_size_bytes,
            compaction_audit.input_file_count,
            compaction_audit.estimated_in_memory_size_bytes_during_discovery,
            compaction_audit.uniform_deltas_created,
            compaction_audit.delta_discovery_time_in_seconds,
            compaction_audit.hash_bucket_time_in_seconds,
            compaction_audit.hash_bucket_invoke_time_in_seconds,
            compaction_audit.hash_bucket_result_size_bytes,
            compaction_audit.total_cluster_object_store_memory_bytes,
            compaction_audit.total_cluster_memory_bytes,
            compaction_audit.total_object_store_memory_used_bytes,
            compaction_audit.hash_bucket_post_object_store_memory_used_bytes,
            compaction_audit.hash_bucket_result_wait_time_in_seconds,
            compaction_audit.peak_memory_used_bytes_per_hash_bucket_task,
            compaction_audit.hash_bucket_processed_size_bytes,
            compaction_audit.input_records,
            compaction_audit.merge_time_in_seconds,
            compaction_audit.merge_invoke_time_in_seconds,
            compaction_audit.merge_result_size,
            compaction_audit.merge_post_object_store_memory_used_bytes,
            compaction_audit.merge_result_wait_time_in_seconds,
            compaction_audit.peak_memory_used_bytes_per_merge_task,
            compaction_audit.records_deduped,
            compaction_audit.records_deleted,
            compaction_audit.compaction_time_in_seconds,
            compaction_audit.peak_memory_used_bytes_by_compaction_session_process,
            compaction_audit.untouched_file_count,
            compaction_audit.untouched_file_ratio,
            compaction_audit.untouched_record_count,
            compaction_audit.untouched_size_bytes,
            compaction_audit.output_file_count,
            compaction_audit.output_size_bytes,
            compaction_audit.output_size_pyarrow_bytes,
            compaction_audit.peak_memory_used_bytes_per_task,
            compaction_audit.pyarrow_version,
            compaction_audit.telemetry_time_in_seconds,
            compaction_audit.observed_input_inflation,
            compaction_audit.observed_input_average_record_size_bytes,
        ]
        for entry in audit_entries:
            assert entry is not None
    return True


def assert_compaction_audit_no_hash_bucket(
    compactor_version: CompactorVersion, compaction_audit: CompactionSessionAuditInfo
):
    if compactor_version == CompactorVersion.V2:
        audit_entries = [
            compaction_audit.deltacat_version,
            compaction_audit.ray_version,
            compaction_audit.audit_url,
            compaction_audit.hash_bucket_count,
            compaction_audit.compactor_version,
            compaction_audit.input_size_bytes,
            compaction_audit.input_file_count,
            compaction_audit.estimated_in_memory_size_bytes_during_discovery,
            compaction_audit.uniform_deltas_created,
            compaction_audit.delta_discovery_time_in_seconds,
            compaction_audit.total_cluster_object_store_memory_bytes,
            compaction_audit.total_cluster_memory_bytes,
            compaction_audit.total_object_store_memory_used_bytes,
            compaction_audit.input_records,
            compaction_audit.merge_time_in_seconds,
            compaction_audit.merge_invoke_time_in_seconds,
            compaction_audit.merge_result_size,
            compaction_audit.merge_post_object_store_memory_used_bytes,
            compaction_audit.merge_result_wait_time_in_seconds,
            compaction_audit.peak_memory_used_bytes_per_merge_task,
            compaction_audit.records_deduped,
            compaction_audit.records_deleted,
            compaction_audit.compaction_time_in_seconds,
            compaction_audit.peak_memory_used_bytes_by_compaction_session_process,
            compaction_audit.untouched_file_count,
            compaction_audit.untouched_file_ratio,
            compaction_audit.untouched_record_count,
            compaction_audit.untouched_size_bytes,
            compaction_audit.output_file_count,
            compaction_audit.output_size_bytes,
            compaction_audit.output_size_pyarrow_bytes,
            compaction_audit.peak_memory_used_bytes_per_task,
            compaction_audit.pyarrow_version,
            compaction_audit.telemetry_time_in_seconds,
        ]
        for entry in audit_entries:
            assert entry is not None
    return True



import logging
import multiprocessing
from functools import partial
from typing import Any, Callable, Dict, Generator, List, Optional, Union
from uuid import uuid4
from botocore.config import Config
from deltacat.aws.constants import (
    BOTO_MAX_RETRIES,
    UPLOAD_DOWNLOAD_RETRY_STOP_AFTER_DELAY,
    BOTO_THROTTLING_ERROR_CODES,
    RETRYABLE_TRANSIENT_ERRORS,
    BOTO_TIMEOUT_ERROR_CODES,
    UPLOAD_SLICED_TABLE_RETRY_STOP_AFTER_DELAY,
    DOWNLOAD_MANIFEST_ENTRY_RETRY_STOP_AFTER_DELAY,
)

import pyarrow as pa
import ray
import s3fs
from boto3.resources.base import ServiceResource
from botocore.client import BaseClient
from botocore.exceptions import ClientError
from ray.data.block import Block, BlockAccessor, BlockMetadata
from ray.data.datasource import FilenameProvider
from ray.types import ObjectRef
from tenacity import (
    Retrying,
    retry_if_exception_type,
    stop_after_delay,
    wait_random_exponential,
)
from deltacat.utils.ray_utils.concurrency import invoke_parallel
import deltacat.aws.clients as aws_utils
from deltacat import logs
from deltacat.storage import (
    DistributedDataset,
    LocalDataset,
    LocalTable,
    Manifest,
    ManifestEntry,
    ManifestEntryList,
)
from deltacat.types.media import (
    ContentEncoding,
    ContentType,
    TableType,
    DistributedDatasetType,
)
from deltacat.types.tables import (
    TABLE_CLASS_TO_SIZE_FUNC,
    TABLE_TYPE_TO_READER_FUNC,
    TABLE_TYPE_TO_DATASET_CREATE_FUNC_REFS,
    DISTRIBUTED_DATASET_TYPE_TO_READER_FUNC,
    get_table_length,
)
from deltacat.exceptions import (
    RetryableError,
    RetryableUploadTableError,
    RetryableDownloadTableError,
    RetryableDownloadFileError,
    RetryableUploadFileError,
    NonRetryableDownloadFileError,
    NonRetryableUploadFileError,
    NonRetryableUploadTableError,
    NonRetryableDownloadTableError,
)
from deltacat.types.partial_download import PartialFileDownloadParams
from deltacat.utils.common import ReadKwargsProvider
from deltacat.exceptions import categorize_errors

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


class CapturedBlockWritePaths:
    def __init__(self):
        self._write_paths: List[str] = []
        self._block_refs: List[ObjectRef[Block]] = []

    def extend(self, write_paths: List[str], block_refs: List[ObjectRef[Block]]):
        try:
            iter(write_paths)
        except TypeError:
            pass
        else:
            self._write_paths.extend(write_paths)
        try:
            iter(block_refs)
        except TypeError:
            pass
        else:
            self._block_refs.extend(block_refs)

    def write_paths(self) -> List[str]:
        return self._write_paths

    def block_refs(self) -> List[ObjectRef[Block]]:
        return self._block_refs


class UuidBlockWritePathProvider(FilenameProvider):
    """Block write path provider implementation that writes each
    dataset block out to a file of the form: {base_path}/{uuid}
    """

    def __init__(
        self, capture_object: CapturedBlockWritePaths, base_path: Optional[str] = None
    ):
        self.base_path = base_path
        self.write_paths: List[str] = []
        self.block_refs: List[ObjectRef[Block]] = []
        self.capture_object = capture_object

    def __del__(self):
        if self.write_paths or self.block_refs:
            self.capture_object.extend(
                self.write_paths,
                self.block_refs,
            )

    def get_filename_for_block(
        self, block: Any, task_index: int, block_index: int
    ) -> str:
        if self.base_path is None:
            raise ValueError(
                "Base path must be provided to UuidBlockWritePathProvider",
            )
        return self._get_write_path_for_block(
            base_path=self.base_path,
            block=block,
            block_index=block_index,
        )

    def _get_write_path_for_block(
        self,
        base_path: str,
        *,
        filesystem: Optional[pa.filesystem.FileSystem] = None,
        dataset_uuid: Optional[str] = None,
        block: Optional[ObjectRef[Block]] = None,
        block_index: Optional[int] = None,
        file_format: Optional[str] = None,
    ) -> str:
        write_path = f"{base_path}/{str(uuid4())}"
        self.write_paths.append(write_path)
        if block:
            self.block_refs.append(block)
        return write_path

    def __call__(
        self,
        base_path: str,
        *,
        filesystem: Optional[pa.filesystem.FileSystem] = None,
        dataset_uuid: Optional[str] = None,
        block: Optional[ObjectRef[Block]] = None,
        block_index: Optional[int] = None,
        file_format: Optional[str] = None,
    ) -> str:
        return self._get_write_path_for_block(
            base_path,
            filesystem=filesystem,
            dataset_uuid=dataset_uuid,
            block=block,
            block_index=block_index,
            file_format=file_format,
        )


class S3Url:
    def __init__(self, url: str):

        from urllib.parse import urlparse

        self._parsed = urlparse(url, allow_fragments=False)  # support '#' in path
        if not self._parsed.scheme:  # support paths w/o 's3://' scheme
            url = f"s3://{url}"
            self._parsed = urlparse(url, allow_fragments=False)
        if self._parsed.query:  # support '?' in path
            self.key = f"{self._parsed.path.lstrip('/')}?{self._parsed.query}"
        else:
            self.key = self._parsed.path.lstrip("/")
        self.bucket = self._parsed.netloc
        self.url = self._parsed.geturl()


def parse_s3_url(url: str) -> S3Url:
    return S3Url(url)


def s3_resource_cache(region: Optional[str], **kwargs) -> ServiceResource:

    return aws_utils.resource_cache(
        "s3",
        region,
        **kwargs,
    )


def s3_client_cache(region: Optional[str], **kwargs) -> BaseClient:

    return aws_utils.client_cache("s3", region, **kwargs)


def get_object_at_url(url: str, **s3_client_kwargs) -> Dict[str, Any]:

    s3 = s3_client_cache(None, **s3_client_kwargs)

    parsed_s3_url = parse_s3_url(url)
    return s3.get_object(Bucket=parsed_s3_url.bucket, Key=parsed_s3_url.key)


def delete_files_by_prefix(bucket: str, prefix: str, **s3_client_kwargs) -> None:

    s3 = s3_resource_cache(None, **s3_client_kwargs)
    bucket = s3.Bucket(bucket)
    bucket.objects.filter(Prefix=prefix).delete()


def filter_paths_by_prefix(bucket, prefix):
    return objects_to_paths(
        bucket,
        filter_objects_by_prefix(bucket, prefix),
    )


def objects_to_paths(bucket, objects):
    for obj in objects:
        yield get_path_from_object(bucket, obj)


def get_path_from_object(bucket, obj):
    return "s3://{}/{}".format(bucket, obj["Key"])


def filter_objects_by_prefix(
    bucket: str, prefix: str, **s3_client_kwargs
) -> Generator[Dict[str, Any], None, None]:

    s3 = s3_client_cache(None, **s3_client_kwargs)
    params = {"Bucket": bucket, "Prefix": prefix}
    more_objects_to_list = True
    while more_objects_to_list:
        response = s3.list_objects_v2(**params)
        if "Contents" in response:
            for obj in response["Contents"]:
                yield obj
        params["ContinuationToken"] = response.get("NextContinuationToken")
        more_objects_to_list = params["ContinuationToken"] is not None


@categorize_errors
def read_file(
    s3_url: str,
    content_type: ContentType,
    content_encoding: ContentEncoding = ContentEncoding.IDENTITY,
    table_type: TableType = TableType.PYARROW,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    partial_file_download_params: Optional[PartialFileDownloadParams] = None,
    **s3_client_kwargs,
) -> LocalTable:

    reader = TABLE_TYPE_TO_READER_FUNC[table_type.value]
    try:
        table = reader(
            s3_url,
            content_type.value,
            content_encoding.value,
            column_names,
            include_columns,
            file_reader_kwargs_provider,
            partial_file_download_params,
            **s3_client_kwargs,
        )
        return table
    except ClientError as e:
        if (
            e.response["Error"]["Code"]
            in BOTO_TIMEOUT_ERROR_CODES | BOTO_THROTTLING_ERROR_CODES
        ):
            # Timeout error not caught by botocore
            raise RetryableDownloadTableError(
                f"Retry table download from: {s3_url} after receiving {type(e).__name__}",
            ) from e
        raise NonRetryableDownloadTableError(
            f"Failed table download from: {s3_url} after receiving {type(e).__name__}"
        ) from e
    except RETRYABLE_TRANSIENT_ERRORS as e:
        raise RetryableDownloadTableError(
            f"Retry download for: {s3_url} after receiving {type(e).__name__}"
        ) from e
    except BaseException as e:
        logger.warn(
            f"Read has failed for {s3_url} and content_type={content_type} "
            f"and encoding={content_encoding}. Error: {e}",
            exc_info=True,
        )
        raise NonRetryableDownloadTableError(
            f"Read has failed for {s3_url} and content_type={content_type} "
            f"and encoding={content_encoding}",
        ) from e


def upload_sliced_table(
    table: Union[LocalTable, DistributedDataset],
    s3_url_prefix: str,
    s3_file_system: s3fs.S3FileSystem,
    max_records_per_entry: Optional[int],
    s3_table_writer_func: Callable,
    table_slicer_func: Callable,
    s3_table_writer_kwargs: Optional[Dict[str, Any]] = None,
    content_type: ContentType = ContentType.PARQUET,
    **s3_client_kwargs,
) -> ManifestEntryList:

    # @retry decorator can't be pickled by Ray, so wrap upload in Retrying
    retrying = Retrying(
        wait=wait_random_exponential(multiplier=1, max=60),
        stop=stop_after_delay(UPLOAD_SLICED_TABLE_RETRY_STOP_AFTER_DELAY),
        retry=retry_if_exception_type(RetryableError),
    )

    manifest_entries = ManifestEntryList()
    table_record_count = get_table_length(table)

    if max_records_per_entry is None or not table_record_count:
        # write the whole table to a single s3 file
        manifest_entries = retrying(
            upload_table,
            table,
            f"{s3_url_prefix}",
            s3_file_system,
            s3_table_writer_func,
            s3_table_writer_kwargs,
            content_type,
            **s3_client_kwargs,
        )
    else:
        # iteratively write table slices
        table_slices = table_slicer_func(table, max_records_per_entry)
        for table_slice in table_slices:
            slice_entries = retrying(
                upload_table,
                table_slice,
                f"{s3_url_prefix}",
                s3_file_system,
                s3_table_writer_func,
                s3_table_writer_kwargs,
                content_type,
                **s3_client_kwargs,
            )
            manifest_entries.extend(slice_entries)
    return manifest_entries


def upload_table(
    table: Union[LocalTable, DistributedDataset],
    s3_base_url: str,
    s3_file_system: s3fs.S3FileSystem,
    s3_table_writer_func: Callable,
    s3_table_writer_kwargs: Optional[Dict[str, Any]],
    content_type: ContentType = ContentType.PARQUET,
    **s3_client_kwargs,
) -> ManifestEntryList:
    """
    Writes the given table to 1 or more S3 files and return Redshift
    manifest entries describing the uploaded files.
    """
    if s3_table_writer_kwargs is None:
        s3_table_writer_kwargs = {}

    capture_object = CapturedBlockWritePaths()
    block_write_path_provider = UuidBlockWritePathProvider(capture_object)
    s3_table_writer_func(
        table,
        s3_base_url,
        s3_file_system,
        block_write_path_provider,
        content_type.value,
        **s3_table_writer_kwargs,
    )
    # TODO: Add a proper fix for block_refs and write_paths not persisting in Ray actors
    del block_write_path_provider
    block_refs = capture_object.block_refs()
    write_paths = capture_object.write_paths()
    metadata = _get_metadata(table, write_paths, block_refs)
    manifest_entries = ManifestEntryList()
    for block_idx, s3_url in enumerate(write_paths):
        try:
            manifest_entry = ManifestEntry.from_s3_obj_url(
                s3_url,
                metadata[block_idx].num_rows,
                metadata[block_idx].size_bytes,
                **s3_client_kwargs,
            )
            manifest_entries.append(manifest_entry)
        except ClientError as e:
            if e.response["Error"]["Code"] == "NoSuchKey":
                # s3fs may swallow S3 errors - we were probably throttled
                raise RetryableUploadTableError(
                    f"Retry table upload from: {s3_url} after receiving {type(e).__name__}",
                ) from e
            if (
                e.response["Error"]["Code"]
                in BOTO_TIMEOUT_ERROR_CODES | BOTO_THROTTLING_ERROR_CODES
            ):
                raise RetryableUploadTableError(
                    f"Retry table upload from: {s3_url} after receiving {type(e).__name__}",
                ) from e
            raise NonRetryableUploadTableError(
                f"Failed table upload to: {s3_url} after receiving {type(e).__name__}",
            ) from e
        except RETRYABLE_TRANSIENT_ERRORS as e:
            raise RetryableUploadTableError(
                f"Retry upload for: {s3_url} after receiving {type(e).__name__}",
            ) from e
        except BaseException as e:
            logger.warn(
                f"Upload has failed for {s3_url} and content_type={content_type}. Error: {e}",
                exc_info=True,
            )
            raise NonRetryableUploadTableError(
                f"Upload has failed for {s3_url} and content_type={content_type} because of {type(e).__name__}",
            ) from e
    return manifest_entries


def download_manifest_entry(
    manifest_entry: ManifestEntry,
    token_holder: Optional[Dict[str, Any]] = None,
    table_type: TableType = TableType.PYARROW,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    content_type: Optional[ContentType] = None,
    content_encoding: Optional[ContentEncoding] = None,
) -> LocalTable:

    s3_client_kwargs = _get_s3_client_kwargs_from_token(token_holder=token_holder)
    if not content_type:
        content_type = manifest_entry.meta.content_type
        assert (
            content_type
        ), f"Unknown content type for manifest entry: {manifest_entry}"
        content_type = ContentType(content_type)
    if not content_encoding:
        content_encoding = manifest_entry.meta.content_encoding
        assert (
            content_encoding
        ), f"Unknown content encoding for manifest entry: {manifest_entry}"
        content_encoding = ContentEncoding(content_encoding)
    s3_url = manifest_entry.uri
    if s3_url is None:
        s3_url = manifest_entry.url

    partial_file_download_params = None
    if manifest_entry.meta and manifest_entry.meta.content_type_parameters:
        for type_params in manifest_entry.meta.content_type_parameters:
            if isinstance(type_params, PartialFileDownloadParams):
                partial_file_download_params = type_params
                break

    # @retry decorator can't be pickled by Ray, so wrap download in Retrying
    retrying = Retrying(
        wait=wait_random_exponential(multiplier=1, max=60),
        stop=stop_after_delay(DOWNLOAD_MANIFEST_ENTRY_RETRY_STOP_AFTER_DELAY),
        retry=retry_if_exception_type(RetryableError),
    )
    table = retrying(
        read_file,
        s3_url,
        content_type,
        content_encoding,
        table_type,
        column_names,
        include_columns,
        file_reader_kwargs_provider,
        partial_file_download_params,
        **s3_client_kwargs,
    )
    return table


@ray.remote
def download_manifest_entry_ray(*args, **kwargs) -> ObjectRef[LocalTable]:
    return download_manifest_entry(*args, **kwargs)


def download_manifest_entries(
    manifest: Manifest,
    token_holder: Optional[Dict[str, Any]] = None,
    table_type: TableType = TableType.PYARROW,
    max_parallelism: Optional[int] = 1,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
) -> LocalDataset:

    if max_parallelism and max_parallelism <= 1:
        return _download_manifest_entries(
            manifest,
            token_holder,
            table_type,
            column_names,
            include_columns,
            file_reader_kwargs_provider,
        )
    else:
        return _download_manifest_entries_parallel(
            manifest,
            token_holder,
            table_type,
            max_parallelism,
            column_names,
            include_columns,
            file_reader_kwargs_provider,
        )


def download_manifest_entries_distributed(
    manifest: Manifest,
    token_holder: Optional[Dict[str, Any]] = None,
    table_type: TableType = TableType.PYARROW,
    max_parallelism: Optional[int] = 1000,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    ray_options_provider: Callable[[int, Any], Dict[str, Any]] = None,
    distributed_dataset_type: Optional[
        DistributedDatasetType
    ] = DistributedDatasetType.RAY_DATASET,
) -> DistributedDataset:

    params = {
        "manifest": manifest,
        "token_holder": token_holder,
        "table_type": table_type,
        "max_parallelism": max_parallelism,
        "column_names": column_names,
        "include_columns": include_columns,
        "file_reader_kwargs_provider": file_reader_kwargs_provider,
        "ray_options_provider": ray_options_provider,
        "distributed_dataset_type": distributed_dataset_type,
    }

    if distributed_dataset_type == DistributedDatasetType.RAY_DATASET:
        return _download_manifest_entries_ray_data_distributed(**params)
    elif distributed_dataset_type is not None:
        return _download_manifest_entries_all_dataset_distributed(**params)
    else:
        raise ValueError(
            f"Distributed dataset type {distributed_dataset_type} not supported."
        )


def upload(s3_url: str, body, **s3_client_kwargs) -> Dict[str, Any]:

    parsed_s3_url = parse_s3_url(s3_url)
    s3 = s3_client_cache(None, **s3_client_kwargs)
    retrying = Retrying(
        wait=wait_random_exponential(multiplier=1, max=15),
        stop=stop_after_delay(UPLOAD_DOWNLOAD_RETRY_STOP_AFTER_DELAY),
        retry=retry_if_exception_type(RetryableError),
    )
    return retrying(
        _put_object,
        s3,
        body,
        parsed_s3_url.bucket,
        parsed_s3_url.key,
    )


def _put_object(
    s3_client, body: Any, bucket: str, key: str, **s3_put_object_kwargs
) -> Dict[str, Any]:
    try:
        return s3_client.put_object(
            Body=body, Bucket=bucket, Key=key, **s3_put_object_kwargs
        )
    except ClientError as e:
        if e.response["Error"]["Code"] in BOTO_THROTTLING_ERROR_CODES:
            error_code = e.response["Error"]["Code"]
            raise RetryableUploadFileError(
                f"Retry upload for: {bucket}/{key} after receiving {error_code}",
            ) from e
        raise NonRetryableUploadFileError(
            f"Failed table upload to: {bucket}/{key}"
        ) from e
    except RETRYABLE_TRANSIENT_ERRORS as e:
        raise RetryableUploadFileError(
            f"Retry upload for: {bucket}/{key} after receiving {type(e).__name__}"
        ) from e
    except BaseException as e:
        logger.error(
            f"Upload has failed for {bucket}/{key}. Error: {type(e).__name__}",
            exc_info=True,
        )
        raise NonRetryableUploadFileError(
            f"Failed table upload to: {bucket}/{key}"
        ) from e


def download(
    s3_url: str, fail_if_not_found: bool = True, **s3_client_kwargs
) -> Optional[Dict[str, Any]]:

    parsed_s3_url = parse_s3_url(s3_url)
    s3 = s3_client_cache(None, **s3_client_kwargs)
    retrying = Retrying(
        wait=wait_random_exponential(multiplier=1, max=15),
        stop=stop_after_delay(UPLOAD_DOWNLOAD_RETRY_STOP_AFTER_DELAY),
        retry=retry_if_exception_type(RetryableError),
    )
    return retrying(
        _get_object,
        s3,
        parsed_s3_url.bucket,
        parsed_s3_url.key,
        fail_if_not_found=fail_if_not_found,
    )


def _get_object(s3_client, bucket: str, key: str, fail_if_not_found: bool = True):
    try:
        return s3_client.get_object(
            Bucket=bucket,
            Key=key,
        )
    except ClientError as e:
        if e.response["Error"]["Code"] == "NoSuchKey":
            if fail_if_not_found:
                raise NonRetryableDownloadFileError(
                    f"Failed get object from: {bucket}/{key}"
                ) from e
            logger.info(f"file not found: {bucket}/{key}")
    except RETRYABLE_TRANSIENT_ERRORS as e:
        raise RetryableDownloadFileError(
            f"Retry get object: {bucket}/{key} after receiving {type(e).__name__}"
        ) from e

    return None


def _download_manifest_entries_parallel(
    manifest: Manifest,
    token_holder: Optional[Dict[str, Any]] = None,
    table_type: TableType = TableType.PYARROW,
    max_parallelism: Optional[int] = None,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
) -> LocalDataset:

    tables = []
    pool = multiprocessing.Pool(max_parallelism)
    downloader = partial(
        download_manifest_entry,
        token_holder=token_holder,
        table_type=table_type,
        column_names=column_names,
        include_columns=include_columns,
        file_reader_kwargs_provider=file_reader_kwargs_provider,
    )
    for table in pool.map(downloader, [e for e in manifest.entries]):
        tables.append(table)
    return tables


def _download_manifest_entries(
    manifest: Manifest,
    token_holder: Optional[Dict[str, Any]] = None,
    table_type: TableType = TableType.PYARROW,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
) -> LocalDataset:

    return [
        download_manifest_entry(
            manifest_entry=e,
            token_holder=token_holder,
            table_type=table_type,
            column_names=column_names,
            include_columns=include_columns,
            file_reader_kwargs_provider=file_reader_kwargs_provider,
        )
        for e in manifest.entries
    ]


@ray.remote
def _block_metadata(block: Block) -> BlockMetadata:
    return BlockAccessor.for_block(block).get_metadata(
        input_files=None,
        exec_stats=None,
    )


def _get_s3_client_kwargs_from_token(token_holder) -> Dict[Any, Any]:
    conf = Config(retries={"max_attempts": BOTO_MAX_RETRIES, "mode": "adaptive"})
    return (
        {
            "aws_access_key_id": token_holder["accessKeyId"],
            "aws_secret_access_key": token_holder["secretAccessKey"],
            "aws_session_token": token_holder["sessionToken"],
            "config": conf,
        }
        if token_holder
        else {"config": conf}
    )


def _get_metadata(
    table: Union[LocalTable, DistributedDataset],
    write_paths: List[str],
    block_refs: List[ObjectRef[Block]],
) -> List[BlockMetadata]:
    metadata: List[BlockMetadata] = []
    if not block_refs:
        # this must be a local table - ensure it was written to only 1 file
        assert len(write_paths) == 1, (
            f"Expected table of type '{type(table)}' to be written to 1 "
            f"file, but found {len(write_paths)} files."
        )
        table_size = None
        table_size_func = TABLE_CLASS_TO_SIZE_FUNC.get(type(table))
        if table_size_func:
            table_size = table_size_func(table)
        else:
            logger.warning(f"Unable to estimate '{type(table)}' table size.")
        metadata.append(
            BlockMetadata(
                num_rows=get_table_length(table),
                size_bytes=table_size,
                schema=None,
                input_files=None,
                exec_stats=None,
            )
        )
    else:
        # TODO(pdames): Expose BlockList metadata getter from Ray Dataset?
        # ray 1.10
        # metadata = dataset._blocks.get_metadata()
        # ray 2.0.0dev
        metadata = table._plan.execute().get_metadata()
        if (
            not metadata
            or metadata[0].size_bytes is None
            or metadata[0].num_rows is None
        ):
            metadata_futures = [
                _block_metadata.remote(block_ref) for block_ref in block_refs
            ]
            metadata = ray.get(metadata_futures)
    return metadata


def _download_manifest_entries_ray_data_distributed(
    manifest: Manifest,
    token_holder: Optional[Dict[str, Any]] = None,
    table_type: TableType = TableType.PYARROW,
    max_parallelism: Optional[int] = 1000,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    ray_options_provider: Callable[[int, Any], Dict[str, Any]] = None,
) -> DistributedDataset:

    table_pending_ids = []
    manifest_entries = manifest.entries
    if manifest_entries:
        table_pending_ids = invoke_parallel(
            manifest_entries,
            download_manifest_entry_ray,
            token_holder,
            table_type,
            column_names,
            include_columns,
            file_reader_kwargs_provider,
            max_parallelism=max_parallelism,
            options_provider=ray_options_provider,
        )
    return TABLE_TYPE_TO_DATASET_CREATE_FUNC_REFS[table_type](table_pending_ids)


def _download_manifest_entries_all_dataset_distributed(
    manifest: Manifest,
    token_holder: Optional[Dict[str, Any]] = None,
    table_type: TableType = TableType.PYARROW,
    max_parallelism: Optional[int] = 1000,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    ray_options_provider: Callable[[int, Any], Dict[str, Any]] = None,
    distributed_dataset_type: Optional[
        DistributedDatasetType
    ] = DistributedDatasetType.RAY_DATASET,
) -> DistributedDataset:

    entry_content_type = None
    entry_content_encoding = None
    uris = []
    for entry in manifest.entries or []:
        if (
            entry_content_type is not None
            and entry_content_type != entry.meta.content_type
        ):
            raise ValueError(
                f"Mixed content types of ({entry_content_type},"
                f" {entry.meta.content_type}) is not supported."
            )

        if (
            entry_content_encoding is not None
            and entry_content_encoding != entry.meta.content_encoding
        ):
            raise ValueError(
                f"Mixed content encoding of {entry_content_encoding},"
                f" {entry.meta.content_encoding} is not supported."
            )

        entry_content_type = entry.meta.content_type
        entry_content_encoding = entry.meta.content_encoding
        uris.append(entry.uri)

    s3_client_kwargs = _get_s3_client_kwargs_from_token(token_holder=token_holder)

    if distributed_dataset_type in DISTRIBUTED_DATASET_TYPE_TO_READER_FUNC:
        return DISTRIBUTED_DATASET_TYPE_TO_READER_FUNC[distributed_dataset_type.value](
            uris=uris,
            content_type=entry_content_type,
            content_encoding=entry_content_encoding,
            column_names=column_names,
            include_columns=include_columns,
            read_func_kwargs_provider=file_reader_kwargs_provider,
            ray_options_provider=ray_options_provider,
            s3_client_kwargs=s3_client_kwargs,
        )
    else:
        raise ValueError(
            f"Unsupported distributed dataset type={distributed_dataset_type}"
        )



import json
import logging
from typing import Dict, Any
from deltacat import logs
from deltacat.compute.compactor import RoundCompletionInfo
from deltacat.storage import PartitionLocator
from deltacat.aws import s3u as s3_utils
from typing import Optional
from deltacat.utils.metrics import metrics

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def get_round_completion_file_s3_url(
    bucket: str,
    source_partition_locator: PartitionLocator,
    destination_partition_locator: Optional[PartitionLocator] = None,
) -> str:

    base_url = source_partition_locator.path(f"s3://{bucket}")
    if destination_partition_locator:
        base_url = destination_partition_locator.path(
            f"s3://{bucket}/{source_partition_locator.hexdigest()}"
        )

    return f"{base_url}.json"


@metrics
def read_round_completion_file(
    bucket: str,
    source_partition_locator: PartitionLocator,
    destination_partition_locator: Optional[PartitionLocator] = None,
    **s3_client_kwargs: Optional[Dict[str, Any]],
) -> RoundCompletionInfo:

    all_uris = []
    if destination_partition_locator:
        round_completion_file_url_with_destination = get_round_completion_file_s3_url(
            bucket,
            source_partition_locator,
            destination_partition_locator,
        )
        all_uris.append(round_completion_file_url_with_destination)

    # Note: we read from RCF at two different URI for backward
    # compatibility reasons.
    round_completion_file_url_prev = get_round_completion_file_s3_url(
        bucket,
        source_partition_locator,
    )

    all_uris.append(round_completion_file_url_prev)

    round_completion_info = None

    for rcf_uri in all_uris:
        logger.info(f"Reading round completion file from: {rcf_uri}")
        result = s3_utils.download(rcf_uri, False, **s3_client_kwargs)
        if result:
            json_str = result["Body"].read().decode("utf-8")
            round_completion_info = RoundCompletionInfo(json.loads(json_str))
            logger.info(f"Read round completion info: {round_completion_info}")
            break
        else:
            logger.warn(f"Round completion file not present at {rcf_uri}")

    return round_completion_info


@metrics
def write_round_completion_file(
    bucket: Optional[str],
    source_partition_locator: Optional[PartitionLocator],
    destination_partition_locator: Optional[PartitionLocator],
    round_completion_info: RoundCompletionInfo,
    completion_file_s3_url: Optional[str] = None,
    **s3_client_kwargs: Optional[Dict[str, Any]],
) -> str:
    if bucket is None and completion_file_s3_url is None:
        raise AssertionError("Either bucket or completion_file_s3_url must be passed")

    logger.info(f"writing round completion file contents: {round_completion_info}")
    if completion_file_s3_url is None:
        completion_file_s3_url = get_round_completion_file_s3_url(
            bucket,
            source_partition_locator,
            destination_partition_locator,
        )
    logger.info(f"writing round completion file to: {completion_file_s3_url}")
    s3_utils.upload(
        completion_file_s3_url,
        str(json.dumps(round_completion_info)),
        **s3_client_kwargs,
    )
    logger.info(f"round completion file written to: {completion_file_s3_url}")
    return completion_file_s3_url



class InvalidNamespaceError(Exception):
    error_name = "InvalidNamespaceError"


class LocalStorageValidationError(Exception):
    error_name = "LocalStorageValidationError"


class LocalStorageError(Exception):
    error_name = "LocalStorageError"



from __future__ import annotations
from typing import List, Optional, Any
from deltacat.storage.model.transform import Transform

"""
An ordered list of partition values determining the values of
ordered transforms specified in the partition spec.
"""
PartitionValues = List[Any]


class PartitionFilter(dict):
    """
    This class represents a filter for partitions.
    It is used to filter partitions based on certain criteria.
    """

    @staticmethod
    def of(
        partition_values: Optional[PartitionValues] = None,
    ) -> PartitionFilter:
        """
        Creates a new PartitionFilter instance with the specified partition key and value.
        """
        partition_filter = PartitionFilter()
        partition_filter["partitionValues"] = partition_values
        return partition_filter

    @property
    def partition_values(self) -> Optional[PartitionValues]:
        return self.get("partitionValues")


class PartitionSpec(dict):
    """
    This class determines how the underlying entities in the
    hierarchy are partitioned. Stream partitions deltas and
    delta partitions files.
    """

    @staticmethod
    def of(ordered_transforms: List[Transform] = None) -> PartitionSpec:
        partition_spec = PartitionSpec()
        partition_spec.ordered_transforms = ordered_transforms
        return partition_spec

    @property
    def ordered_transforms(self) -> List[Transform]:
        return self.get("orderedTransforms")

    @ordered_transforms.setter
    def ordered_transforms(self, value: List[Transform]) -> None:
        self["orderedTransforms"] = value


class StreamPartitionSpec(PartitionSpec):
    """
    A class representing a stream partition specification.
    A stream partitions deltas into multiple different Partition
    """

    pass


class DeltaPartitionSpec(PartitionSpec):
    """
    A class representing delta partition specification.
    The manifest entries in delta are partitioned based on this spec.
    """

    pass



from enum import Enum
from typing import List, Union

from pyarrow.parquet import ParquetFile
import numpy as np
import pandas as pd
import pyarrow as pa
from ray.data.dataset import Dataset
from daft import DataFrame as DaftDataFrame

LocalTable = Union[pa.Table, pd.DataFrame, np.ndarray, ParquetFile]
LocalDataset = List[LocalTable]
DistributedDataset = Union[Dataset, DaftDataFrame]


class DeltaType(str, Enum):
    APPEND = "append"
    UPSERT = "upsert"
    DELETE = "delete"


class LifecycleState(str, Enum):
    UNRELEASED = "unreleased"
    ACTIVE = "active"
    DEPRECATED = "deprecated"
    BETA = "beta"
    DELETED = "deleted"


class CommitState(str, Enum):
    STAGED = "staged"
    COMMITTED = "committed"
    DEPRECATED = "deprecated"


class SchemaConsistencyType(str, Enum):
    """
    Schemas are optional for DeltaCAT tables and can be used to inform the data
    consistency checks run for each field. If a schema is present, it can be
    used to enforce the following column-level data consistency policies at
    table load time:

    NONE: No consistency checks are run. May be mixed with the below two
    policies by specifying column names to pass through together with
    column names to coerce/validate.

    COERCE: Coerce fields to fit the schema whenever possible. An explicit
    subset of column names to coerce may optionally be specified.

    VALIDATE: Raise an error for any fields that don't fit the schema. An
    explicit subset of column names to validate may optionally be specified.
    """

    NONE = "none"
    COERCE = "coerce"
    VALIDATE = "validate"



from typing import Any, Dict, List, Optional, Set, Union, Tuple
import pyarrow as pa
import logging
from deltacat.catalog.model.table_definition import TableDefinition
from deltacat.storage.model.sort_key import SortKey
from deltacat.storage.model.list_result import ListResult
from deltacat.storage.model.namespace import Namespace
from deltacat.storage.model.types import (
    DistributedDataset,
    LifecycleState,
    LocalDataset,
    LocalTable,
    SchemaConsistencyType,
)
from deltacat.storage.model.partition import PartitionLocator, Partition
from deltacat.storage.model.table_version import TableVersion
from deltacat.compute.merge_on_read.model.merge_on_read_params import MergeOnReadParams
from deltacat.storage.model.delta import DeltaType
import deltacat.storage.interface as deltacat_storage
from deltacat.types.media import ContentType, TableType, DistributedDatasetType
from deltacat.types.tables import TableWriteMode
from deltacat.compute.merge_on_read import MERGE_FUNC_BY_DISTRIBUTED_DATASET_TYPE
from deltacat import logs

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

STORAGE = None


# table functions
def write_to_table(
    data: Union[LocalTable, LocalDataset, DistributedDataset],  # type: ignore
    table: str,
    namespace: Optional[str] = None,
    mode: TableWriteMode = TableWriteMode.AUTO,
    content_type: ContentType = ContentType.PARQUET,
    *args,
    **kwargs,
) -> None:
    """Write local or distributed data to a table. Raises an error if the
    table does not exist and the table write mode is not CREATE or AUTO.

    When creating a table, all `create_table` parameters may be optionally
    specified as additional keyword arguments. When appending to, or replacing,
    an existing table, all `alter_table` parameters may be optionally specified
    as additional keyword arguments."""
    raise NotImplementedError("write_to_table not implemented")


def read_table(
    table: str,
    namespace: Optional[str] = None,
    table_version: Optional[str] = None,
    table_type: Optional[TableType] = TableType.PYARROW,
    distributed_dataset_type: Optional[
        DistributedDatasetType
    ] = DistributedDatasetType.RAY_DATASET,
    partition_filter: Optional[List[Union[Partition, PartitionLocator]]] = None,
    stream_position_range_inclusive: Optional[Tuple[int, int]] = None,
    merge_on_read: Optional[bool] = False,
    reader_kwargs: Optional[Dict[Any, Any]] = None,
    deltacat_storage_kwargs: Optional[Dict[Any, Any]] = None,
    *args,
    **kwargs,
) -> DistributedDataset:  # type: ignore
    """Read a table into a distributed dataset."""

    if reader_kwargs is None:
        reader_kwargs = {}

    if deltacat_storage_kwargs is None:
        deltacat_storage_kwargs = {}

    _validate_read_table_args(
        namespace=namespace,
        table_type=table_type,
        distributed_dataset_type=distributed_dataset_type,
        merge_on_read=merge_on_read,
    )

    table_version_obj = _get_latest_or_given_table_version(
        namespace=namespace,
        table_name=table,
        table_version=table_version,
        **deltacat_storage_kwargs,
    )
    table_version = table_version_obj.table_version

    if (
        table_version_obj.content_types is None
        or len(table_version_obj.content_types) != 1
    ):
        raise ValueError(
            "Expected exactly one content type but "
            f"found {table_version_obj.content_types}."
        )

    logger.info(
        f"Reading metadata for table={namespace}/{table}/{table_version} "
        f"with partition_filters={partition_filter} and stream position"
        f" range={stream_position_range_inclusive}"
    )

    if partition_filter is None:
        logger.info(
            f"Reading all partitions metadata in the table={table} "
            "as partition_filter was None."
        )
        partition_filter = STORAGE.list_partitions(
            table_name=table,
            namespace=namespace,
            table_version=table_version,
            **deltacat_storage_kwargs,
        ).all_items()

    qualified_deltas = _get_deltas_from_partition_filter(
        stream_position_range_inclusive=stream_position_range_inclusive,
        partition_filter=partition_filter,
        **deltacat_storage_kwargs,
    )

    logger.info(
        f"Total qualified deltas={len(qualified_deltas)} "
        f"from {len(partition_filter)} partitions."
    )

    merge_on_read_params = MergeOnReadParams.of(
        {
            "deltas": qualified_deltas,
            "deltacat_storage": STORAGE,
            "deltacat_storage_kwargs": deltacat_storage_kwargs,
            "reader_kwargs": reader_kwargs,
        }
    )

    return MERGE_FUNC_BY_DISTRIBUTED_DATASET_TYPE[distributed_dataset_type.value](
        params=merge_on_read_params, **kwargs
    )


def alter_table(
    table: str,
    namespace: Optional[str] = None,
    lifecycle_state: Optional[LifecycleState] = None,
    schema_updates: Optional[Dict[str, Any]] = None,
    partition_updates: Optional[Dict[str, Any]] = None,
    primary_keys: Optional[Set[str]] = None,
    sort_keys: Optional[List[SortKey]] = None,
    description: Optional[str] = None,
    properties: Optional[Dict[str, str]] = None,
    *args,
    **kwargs,
) -> None:
    """Alter table definition."""
    raise NotImplementedError("alter_table not implemented")


def create_table(
    table: str,
    namespace: Optional[str] = None,
    lifecycle_state: Optional[LifecycleState] = None,
    schema: Optional[Union[pa.Schema, str, bytes]] = None,
    schema_consistency: Optional[Dict[str, SchemaConsistencyType]] = None,
    partition_keys: Optional[List[Dict[str, Any]]] = None,
    primary_keys: Optional[Set[str]] = None,
    sort_keys: Optional[List[SortKey]] = None,
    description: Optional[str] = None,
    properties: Optional[Dict[str, str]] = None,
    permissions: Optional[Dict[str, Any]] = None,
    content_types: Optional[List[ContentType]] = None,
    replace_existing_table: bool = False,
    *args,
    **kwargs,
) -> TableDefinition:
    """Create an empty table. Raises an error if the table already exists and
    `replace_existing_table` is False."""
    raise NotImplementedError("create_table not implemented")


def drop_table(
    table: str, namespace: Optional[str] = None, purge: bool = False, *args, **kwargs
) -> None:
    """Drop a table from the catalog and optionally purge it. Raises an error
    if the table does not exist."""
    raise NotImplementedError("drop_table not implemented")


def refresh_table(table: str, namespace: Optional[str] = None, *args, **kwargs) -> None:
    """Refresh metadata cached on the Ray cluster for the given table."""
    raise NotImplementedError("refresh_table not implemented")


def list_tables(
    namespace: Optional[str] = None, *args, **kwargs
) -> ListResult[TableDefinition]:
    """List a page of table definitions. Raises an error if the given namespace
    does not exist."""
    raise NotImplementedError("list_tables not implemented")


def get_table(
    table: str, namespace: Optional[str] = None, *args, **kwargs
) -> Optional[TableDefinition]:
    """Get table definition metadata. Returns None if the given table does not
    exist."""
    raise NotImplementedError("get_table not implemented")


def truncate_table(
    table: str, namespace: Optional[str] = None, *args, **kwargs
) -> None:
    """Truncate table data. Raises an error if the table does not exist."""
    raise NotImplementedError("truncate_table not implemented")


def rename_table(
    table: str, new_name: str, namespace: Optional[str] = None, *args, **kwargs
) -> None:
    """Rename a table."""
    raise NotImplementedError("rename_table not implemented")


def table_exists(table: str, namespace: Optional[str] = None, *args, **kwargs) -> bool:
    """Returns True if the given table exists, False if not."""
    raise NotImplementedError("table_exists not implemented")


# namespace functions
def list_namespaces(*args, **kwargs) -> ListResult[Namespace]:
    """List a page of table namespaces."""
    raise NotImplementedError("list_namespaces not implemented")


def get_namespace(namespace: str, *args, **kwargs) -> Optional[Namespace]:
    """Gets table namespace metadata for the specified table namespace. Returns
    None if the given namespace does not exist."""
    raise NotImplementedError("get_namespace not implemented")


def namespace_exists(namespace: str, *args, **kwargs) -> bool:
    """Returns True if the given table namespace exists, False if not."""
    raise NotImplementedError("namespace_exists not implemented")


def create_namespace(
    namespace: str, permissions: Dict[str, Any], *args, **kwargs
) -> Namespace:
    """Creates a table namespace with the given name and permissions. Returns
    the created namespace. Raises an error if the namespace already exists."""
    raise NotImplementedError("create_namespace not implemented")


def alter_namespace(
    namespace: str,
    permissions: Optional[Dict[str, Any]] = None,
    new_namespace: Optional[str] = None,
    *args,
    **kwargs,
) -> None:
    """Alter table namespace definition."""
    raise NotImplementedError("alter_namespace not implemented")


def drop_namespace(namespace: str, purge: bool = False, *args, **kwargs) -> None:
    """Drop the given namespace and all of its tables from the catalog,
    optionally purging them."""
    raise NotImplementedError("drop_namespace not implemented")


def default_namespace() -> str:
    """Returns the default namespace for the catalog."""
    raise NotImplementedError("default_namespace not implemented")


# catalog functions
def initialize(ds: deltacat_storage, *args, **kwargs) -> None:
    """Initializes the data catalog with the given arguments."""
    global STORAGE
    STORAGE = ds


def _validate_read_table_args(
    namespace: Optional[str] = None,
    table_type: Optional[TableType] = None,
    distributed_dataset_type: Optional[DistributedDatasetType] = None,
    merge_on_read: Optional[bool] = None,
):
    if STORAGE is None:
        raise ValueError(
            "Catalog not initialized. Did you miss calling "
            "initialize(ds=<deltacat_storage>)?"
        )

    if merge_on_read:
        raise ValueError("Merge on read not supported currently.")

    if table_type is not TableType.PYARROW:
        raise ValueError("Only PYARROW table type is supported as of now")

    if distributed_dataset_type is not DistributedDatasetType.DAFT:
        raise ValueError("Only DAFT dataset type is supported as of now")

    if namespace is None:
        raise ValueError(
            "namespace must be passed to uniquely identify a table in the catalog."
        )


def _get_latest_or_given_table_version(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs,
) -> TableVersion:
    table_version_obj = None
    if table_version is None:
        table_version_obj = STORAGE.get_latest_table_version(
            namespace=namespace, table_name=table_name, *args, **kwargs
        )
        table_version = table_version_obj.table_version
    else:
        table_version_obj = STORAGE.get_table_version(
            namespace=namespace,
            table_name=table_name,
            table_version=table_version,
            *args,
            **kwargs,
        )

    return table_version_obj


def _get_deltas_from_partition_filter(
    partition_filter: Optional[List[Union[Partition, PartitionLocator]]] = None,
    stream_position_range_inclusive: Optional[Tuple[int, int]] = None,
    *args,
    **kwargs,
):

    result_deltas = []
    start_stream_position, end_stream_position = stream_position_range_inclusive or (
        None,
        None,
    )
    for partition_like in partition_filter:
        deltas = STORAGE.list_partition_deltas(
            partition_like=partition_like,
            ascending_order=True,
            include_manifest=True,
            start_stream_position=start_stream_position,
            last_stream_position=end_stream_position,
            *args,
            **kwargs,
        ).all_items()

        for delta in deltas:
            if (
                start_stream_position is None
                or delta.stream_position >= start_stream_position
            ) and (
                end_stream_position is None
                or delta.stream_position <= end_stream_position
            ):
                if delta.type == DeltaType.DELETE:
                    raise ValueError("DELETE type deltas are not supported")
                result_deltas.append(delta)

    return result_deltas



import ray
import os
from moto import mock_s3
import pytest
import boto3
from boto3.resources.base import ServiceResource
import pyarrow as pa
from deltacat.io.ray_plasma_object_store import RayPlasmaObjectStore
from pytest_benchmark.fixture import BenchmarkFixture

from deltacat.tests.compute.test_util_constant import (
    TEST_S3_RCF_BUCKET_NAME,
    DEFAULT_NUM_WORKERS,
    DEFAULT_WORKER_INSTANCE_CPUS,
)
from deltacat.tests.compute.test_util_common import (
    get_rcf,
)
from deltacat.tests.test_utils.utils import read_s3_contents
from deltacat.compute.compactor.model.compactor_version import CompactorVersion
from deltacat.tests.compute.test_util_common import (
    get_compacted_delta_locator_from_rcf,
)
from deltacat.compute.compactor.model.compaction_session_audit_info import (
    CompactionSessionAuditInfo,
)
from deltacat.tests.compute.test_util_create_table_deltas_repo import (
    multiple_rounds_create_src_w_deltas_destination_rebase_w_deltas_strategy,
)
from deltacat.tests.compute.compact_partition_multiple_rounds_test_cases import (
    MULTIPLE_ROUNDS_TEST_CASES,
)
from typing import Any, Callable, Dict, List, Optional, Set
from deltacat.types.media import StorageType
from deltacat.storage import (
    DeltaLocator,
    Partition,
)
from deltacat.types.media import ContentType
from deltacat.compute.compactor.model.compact_partition_params import (
    CompactPartitionParams,
)
from deltacat.compute.compactor import (
    RoundCompletionInfo,
)
from deltacat.utils.placement import (
    PlacementGroupManager,
)

DATABASE_FILE_PATH_KEY, DATABASE_FILE_PATH_VALUE = (
    "db_file_path",
    "deltacat/tests/local_deltacat_storage/db_test.sqlite",
)


"""
MODULE scoped fixtures
"""


@pytest.fixture(autouse=True, scope="module")
def setup_ray_cluster():
    ray.init(local_mode=True, ignore_reinit_error=True)
    yield
    ray.shutdown()


@pytest.fixture(autouse=True, scope="module")
def mock_aws_credential():
    os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    os.environ["AWS_SECRET_ACCESS_ID"] = "testing"
    os.environ["AWS_SECURITY_TOKEN"] = "testing"
    os.environ["AWS_SESSION_TOKEN"] = "testing"
    os.environ["AWS_DEFAULT_REGION"] = "us-east-1"
    yield


@pytest.fixture(autouse=True, scope="module")
def cleanup_the_database_file_after_all_compaction_session_package_tests_complete():
    # make sure the database file is deleted after all the compactor package tests are completed
    if os.path.exists(DATABASE_FILE_PATH_VALUE):
        os.remove(DATABASE_FILE_PATH_VALUE)


@pytest.fixture(scope="module")
def s3_resource(mock_aws_credential):
    with mock_s3():
        yield boto3.resource("s3")


@pytest.fixture(autouse=True, scope="module")
def setup_compaction_artifacts_s3_bucket(s3_resource: ServiceResource):
    s3_resource.create_bucket(
        ACL="authenticated-read",
        Bucket=TEST_S3_RCF_BUCKET_NAME,
    )
    yield


"""
FUNCTION scoped fixtures
"""


@pytest.fixture(scope="function")
def local_deltacat_storage_kwargs(request: pytest.FixtureRequest):
    # see deltacat/tests/local_deltacat_storage/README.md for documentation
    kwargs_for_local_deltacat_storage: Dict[str, Any] = {
        DATABASE_FILE_PATH_KEY: DATABASE_FILE_PATH_VALUE,
    }
    yield kwargs_for_local_deltacat_storage
    if os.path.exists(DATABASE_FILE_PATH_VALUE):
        os.remove(DATABASE_FILE_PATH_VALUE)


@pytest.mark.parametrize(
    [
        "test_name",
        "primary_keys",
        "sort_keys",
        "partition_keys_param",
        "partition_values_param",
        "input_deltas_param",
        "expected_terminal_compact_partition_result",
        "expected_terminal_exception",
        "expected_terminal_exception_message",
        "create_placement_group_param",
        "records_per_compacted_file_param",
        "hash_bucket_count_param",
        "read_kwargs_provider_param",
        "drop_duplicates_param",
        "skip_enabled_compact_partition_drivers",
        "assert_compaction_audit",
        "rebase_expected_compact_partition_result",
        "num_rounds_param",
        "compact_partition_func",
        "compactor_version",
    ],
    [
        (
            test_name,
            primary_keys,
            sort_keys,
            partition_keys_param,
            partition_values_param,
            input_deltas,
            expected_terminal_compact_partition_result,
            expected_terminal_exception,
            expected_terminal_exception_message,
            create_placement_group_param,
            records_per_compacted_file_param,
            hash_bucket_count_param,
            drop_duplicates_param,
            read_kwargs_provider,
            skip_enabled_compact_partition_drivers,
            assert_compaction_audit,
            rebase_expected_compact_partition_result,
            num_rounds_param,
            compact_partition_func,
            compactor_version,
        )
        for test_name, (
            primary_keys,
            sort_keys,
            partition_keys_param,
            partition_values_param,
            input_deltas,
            expected_terminal_compact_partition_result,
            expected_terminal_exception,
            expected_terminal_exception_message,
            create_placement_group_param,
            records_per_compacted_file_param,
            hash_bucket_count_param,
            drop_duplicates_param,
            read_kwargs_provider,
            skip_enabled_compact_partition_drivers,
            assert_compaction_audit,
            rebase_expected_compact_partition_result,
            num_rounds_param,
            compact_partition_func,
            compactor_version,
        ) in MULTIPLE_ROUNDS_TEST_CASES.items()
    ],
    ids=[test_name for test_name in MULTIPLE_ROUNDS_TEST_CASES],
)
def test_compact_partition_rebase_multiple_rounds_same_source_and_destination(
    mocker,
    s3_resource: ServiceResource,
    local_deltacat_storage_kwargs: Dict[str, Any],
    test_name: str,
    primary_keys: Set[str],
    sort_keys: List[Optional[Any]],
    partition_keys_param: Optional[List[Any]],
    partition_values_param: List[Optional[str]],
    input_deltas_param: List[pa.Array],
    expected_terminal_compact_partition_result: pa.Table,
    expected_terminal_exception: BaseException,
    expected_terminal_exception_message: Optional[str],
    create_placement_group_param: bool,
    records_per_compacted_file_param: int,
    hash_bucket_count_param: int,
    drop_duplicates_param: bool,
    read_kwargs_provider_param: Any,
    rebase_expected_compact_partition_result: pa.Table,
    skip_enabled_compact_partition_drivers: List[CompactorVersion],
    assert_compaction_audit: Optional[Callable],
    compactor_version: Optional[CompactorVersion],
    compact_partition_func: Callable,
    num_rounds_param: int,
    benchmark: BenchmarkFixture,
):
    import deltacat.tests.local_deltacat_storage as ds

    ds_mock_kwargs = local_deltacat_storage_kwargs
    """
    This test tests different multi-round compaction rebase configurations,
    as specified in compact_partition_multiple_rounds_test_cases.py
    These tests do not test multi-round compaction backfill, which is
    currently unsupported.
    """
    (
        source_table_stream,
        _,
        rebased_table_stream,
        _,
    ) = multiple_rounds_create_src_w_deltas_destination_rebase_w_deltas_strategy(
        primary_keys,
        sort_keys,
        partition_keys_param,
        input_deltas_param,
        partition_values_param,
        ds_mock_kwargs,
    )
    source_partition: Partition = ds.get_partition(
        source_table_stream.locator,
        partition_values_param,
        **ds_mock_kwargs,
    )
    rebased_partition: Partition = ds.get_partition(
        rebased_table_stream.locator,
        partition_values_param,
        **ds_mock_kwargs,
    )
    total_cpus = DEFAULT_NUM_WORKERS * DEFAULT_WORKER_INSTANCE_CPUS
    pgm = None
    if create_placement_group_param:
        pgm = PlacementGroupManager(
            1, total_cpus, DEFAULT_WORKER_INSTANCE_CPUS, memory_per_bundle=4000000
        ).pgs[0]
    compact_partition_params = CompactPartitionParams.of(
        {
            "compaction_artifact_s3_bucket": TEST_S3_RCF_BUCKET_NAME,
            "compacted_file_content_type": ContentType.PARQUET,
            "dd_max_parallelism_ratio": 1.0,
            "deltacat_storage": ds,
            "deltacat_storage_kwargs": ds_mock_kwargs,
            "destination_partition_locator": rebased_partition.locator,
            "hash_bucket_count": hash_bucket_count_param,
            "last_stream_position_to_compact": source_partition.stream_position,
            "list_deltas_kwargs": {**ds_mock_kwargs, **{"equivalent_table_types": []}},
            "object_store": RayPlasmaObjectStore(),
            "pg_config": pgm,
            "primary_keys": primary_keys,
            "read_kwargs_provider": read_kwargs_provider_param,
            "rebase_source_partition_locator": source_partition.locator,
            "rebase_source_partition_high_watermark": rebased_partition.stream_position,
            "records_per_compacted_file": records_per_compacted_file_param,
            "s3_client_kwargs": {},
            "source_partition_locator": rebased_partition.locator,
            "sort_keys": sort_keys if sort_keys else None,
            "num_rounds": num_rounds_param,
            "drop_duplicates": drop_duplicates_param,
            "min_delta_bytes": 560,
        }
    )
    if expected_terminal_exception:
        with pytest.raises(expected_terminal_exception) as exc_info:
            benchmark(compact_partition_func, compact_partition_params)
        assert expected_terminal_exception_message in str(exc_info.value)
        return
    from deltacat.compute.compactor_v2.model.evaluate_compaction_result import (
        ExecutionCompactionResult,
    )

    execute_compaction_result_spy = mocker.spy(ExecutionCompactionResult, "__init__")

    # execute
    rcf_file_s3_uri = benchmark(compact_partition_func, compact_partition_params)

    round_completion_info: RoundCompletionInfo = get_rcf(s3_resource, rcf_file_s3_uri)
    audit_bucket, audit_key = RoundCompletionInfo.get_audit_bucket_name_and_key(
        round_completion_info.compaction_audit_url
    )

    compaction_audit_obj: Dict[str, Any] = read_s3_contents(
        s3_resource, audit_bucket, audit_key
    )
    compaction_audit: CompactionSessionAuditInfo = CompactionSessionAuditInfo(
        **compaction_audit_obj
    )

    # Assert not in-place compacted
    assert (
        execute_compaction_result_spy.call_args.args[-1] is False
    ), "Table version erroneously marked as in-place compacted!"
    compacted_delta_locator: DeltaLocator = get_compacted_delta_locator_from_rcf(
        s3_resource, rcf_file_s3_uri
    )
    tables = ds.download_delta(
        compacted_delta_locator, storage_type=StorageType.LOCAL, **ds_mock_kwargs
    )
    actual_rebase_compacted_table = pa.concat_tables(tables)
    # if no primary key is specified then sort by sort_key for consistent assertion
    sorting_cols: List[Any] = (
        [(val, "ascending") for val in primary_keys] if primary_keys else sort_keys
    )
    rebase_expected_compact_partition_result = (
        rebase_expected_compact_partition_result.combine_chunks().sort_by(sorting_cols)
    )
    actual_rebase_compacted_table = (
        actual_rebase_compacted_table.combine_chunks().sort_by(sorting_cols)
    )
    assert actual_rebase_compacted_table.equals(
        rebase_expected_compact_partition_result
    ), f"{actual_rebase_compacted_table} does not match {rebase_expected_compact_partition_result}"

    if assert_compaction_audit:
        if not assert_compaction_audit(compactor_version, compaction_audit):
            assert False, "Compaction audit assertion failed"
    return



from typing import Any, Callable, Dict, List, Optional, Set, Union, Tuple

import pyarrow as pa
import daft
import json
import sqlite3
from sqlite3 import Cursor, Connection
import uuid
import ray
import io

from deltacat.tests.test_utils.storage import create_empty_delta
from deltacat.utils.common import current_time_ms


from deltacat.storage import (
    Delta,
    DeltaLocator,
    DeltaType,
    DistributedDataset,
    LifecycleState,
    ListResult,
    LocalDataset,
    LocalTable,
    Manifest,
    ManifestAuthor,
    Namespace,
    NamespaceLocator,
    Partition,
    SchemaConsistencyType,
    Stream,
    StreamLocator,
    Table,
    TableVersion,
    TableVersionLocator,
    TableLocator,
    CommitState,
    SortKey,
    PartitionLocator,
    ManifestMeta,
    ManifestEntry,
    ManifestEntryList,
    DeleteParameters,
    PartitionFilter,
    PartitionValues,
    DeltaPartitionSpec,
    StreamPartitionSpec,
    TransformName,
    IdentityTransformParameters,
)
from deltacat.types.media import (
    ContentType,
    StorageType,
    TableType,
    ContentEncoding,
    DistributedDatasetType,
)
from deltacat.utils.common import ReadKwargsProvider
from deltacat.tests.local_deltacat_storage.exceptions import (
    InvalidNamespaceError,
    LocalStorageValidationError,
)

SQLITE_CUR_ARG = "sqlite3_cur"
SQLITE_CON_ARG = "sqlite3_con"
DB_FILE_PATH_ARG = "db_file_path"

STORAGE_TYPE = "SQLITE3"
STREAM_ID_PROPERTY = "stream_id"
CREATE_NAMESPACES_TABLE = (
    "CREATE TABLE IF NOT EXISTS namespaces(locator, value, PRIMARY KEY (locator))"
)
CREATE_TABLES_TABLE = (
    "CREATE TABLE IF NOT EXISTS tables(locator, namespace_locator, value, PRIMARY KEY (locator), "
    "FOREIGN KEY (namespace_locator) REFERENCES namespaces(locator))"
)
CREATE_TABLE_VERSIONS_TABLE = (
    "CREATE TABLE IF NOT EXISTS table_versions(locator, table_locator, value, PRIMARY KEY (locator), "
    "FOREIGN KEY (table_locator) REFERENCES tables(locator))"
)
CREATE_STREAMS_TABLE = (
    "CREATE TABLE IF NOT EXISTS streams(locator, table_version_locator, value, PRIMARY KEY(locator), "
    "FOREIGN KEY (table_version_locator) REFERENCES table_versions(locator))"
)
CREATE_PARTITIONS_TABLE = (
    "CREATE TABLE IF NOT EXISTS partitions(locator, stream_locator, value, PRIMARY KEY(locator), "
    "FOREIGN KEY (stream_locator) REFERENCES streams(locator))"
)
CREATE_DELTAS_TABLE = (
    "CREATE TABLE IF NOT EXISTS deltas(locator, partition_locator, value, PRIMARY KEY(locator), "
    "FOREIGN KEY (partition_locator) REFERENCES partitions(locator))"
)
CREATE_DATA_TABLE = "CREATE TABLE IF NOT EXISTS data(uri, value, PRIMARY KEY(uri))"


def _get_sqlite3_cursor_con(kwargs) -> Tuple[Cursor, Connection]:
    if SQLITE_CUR_ARG in kwargs and SQLITE_CON_ARG in kwargs:
        return kwargs[SQLITE_CUR_ARG], kwargs[SQLITE_CON_ARG]
    elif DB_FILE_PATH_ARG in kwargs:
        con = sqlite3.connect(kwargs[DB_FILE_PATH_ARG])
        cur = con.cursor()
        return cur, con

    raise ValueError(f"Invalid local db connection kwargs: {kwargs}")


def _get_manifest_entry_uri(manifest_entry_id: str) -> str:
    return f"cloudpickle://{manifest_entry_id}"


def _merge_and_promote(
    partition_deltas: List[Delta], previous_partition_deltas: List[Delta]
):
    previous_partition_deltas_spos_gt: List[Delta] = [
        delta
        for delta in previous_partition_deltas
        if delta.stream_position > partition_deltas[0].stream_position
    ]
    # handle the case if the previous partition deltas have a greater stream position than the partition_delta
    partition_deltas = previous_partition_deltas_spos_gt + partition_deltas
    return partition_deltas


def list_namespaces(*args, **kwargs) -> ListResult[Namespace]:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    res = cur.execute("SELECT * FROM namespaces")
    fetched = res.fetchall()
    result = []

    for item in fetched:
        result.append(Namespace(json.loads(item[1])))

    return ListResult.of(result, None, None)


def list_tables(namespace: str, *args, **kwargs) -> ListResult[Table]:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    params = (NamespaceLocator.of(namespace).canonical_string(),)
    res = cur.execute("SELECT * FROM tables WHERE namespace_locator = ?", params)
    fetched = res.fetchall()
    result = []

    for item in fetched:
        result.append(Table(json.loads(item[2])))

    return ListResult.of(result, None, None)


def list_table_versions(
    namespace: str, table_name: str, *args, **kwargs
) -> ListResult[TableVersion]:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    table_locator = TableLocator.of(NamespaceLocator.of(namespace), table_name)

    res = cur.execute(
        "SELECT * FROM table_versions WHERE table_locator = ?",
        (table_locator.canonical_string(),),
    )
    fetched = res.fetchall()
    result = []

    for item in fetched:
        result.append(TableVersion(json.loads(item[2])))

    return ListResult.of(result, None, None)


def list_partitions(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs,
) -> ListResult[Partition]:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    stream = get_stream(namespace, table_name, table_version, *args, **kwargs)

    res = cur.execute(
        "SELECT * FROM partitions WHERE stream_locator = ?",
        (stream.locator.canonical_string(),),
    )

    fetched = res.fetchall()
    result = []
    for item in fetched:
        partition = Partition(json.loads(item[2]))
        if partition.state == CommitState.COMMITTED:
            result.append(partition)

    return ListResult.of(result, None, None)


def list_stream_partitions(stream: Stream, *args, **kwargs) -> ListResult[Partition]:
    return list_partitions(
        stream.namespace, stream.table_name, stream.table_version, *args, **kwargs
    )


def list_deltas(
    namespace: str,
    table_name: str,
    partition_values: Optional[PartitionValues] = None,
    table_version: Optional[str] = None,
    first_stream_position: Optional[int] = None,
    last_stream_position: Optional[int] = None,
    ascending_order: Optional[bool] = None,
    include_manifest: bool = False,
    partition_filter: Optional[PartitionFilter] = None,
    *args,
    **kwargs,
) -> ListResult[Delta]:
    stream = get_stream(namespace, table_name, table_version, *args, **kwargs)
    if stream is None:
        return ListResult.of([], None, None)

    if partition_values is not None and partition_filter is not None:
        raise ValueError(
            "Only one of partition_values or partition_filter must be provided"
        )
    if partition_filter is not None:
        partition_values = partition_filter.partition_values

    partition = get_partition(stream.locator, partition_values, *args, **kwargs)

    all_deltas = list_partition_deltas(
        partition,
        first_stream_position=first_stream_position,
        last_stream_position=last_stream_position,
        ascending_order=ascending_order,
        include_manifest=include_manifest,
        *args,
        **kwargs,
    ).all_items()

    result = []

    for delta in all_deltas:
        if (
            not first_stream_position or first_stream_position < delta.stream_position
        ) and (
            not last_stream_position or delta.stream_position <= last_stream_position
        ):
            result.append(delta)

        if not include_manifest:
            delta.manifest = None

    result.sort(reverse=(not ascending_order), key=lambda d: d.stream_position)
    return ListResult.of(result, None, None)


def list_partition_deltas(
    partition_like: Union[Partition, PartitionLocator],
    first_stream_position: Optional[int] = None,
    last_stream_position: Optional[int] = None,
    ascending_order: bool = False,
    include_manifest: bool = False,
    *args,
    **kwargs,
) -> ListResult[Delta]:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    if partition_like is None:
        return ListResult.of([], None, None)

    if first_stream_position is None:
        first_stream_position = 0

    if last_stream_position is None:
        last_stream_position = float("inf")

    assert isinstance(partition_like, Partition) or isinstance(
        partition_like, PartitionLocator
    ), f"Expected a Partition or PartitionLocator as an input argument but found {partition_like}"

    partition_locator = None
    if isinstance(partition_like, Partition):
        partition_locator = partition_like.locator
    else:
        partition_locator = partition_like

    res = cur.execute(
        "SELECT * FROM deltas WHERE partition_locator = ?",
        (partition_locator.canonical_string(),),
    )

    serialized_items = res.fetchall()

    if not serialized_items:
        return ListResult.of([], None, None)

    result = []
    for item in serialized_items:
        current_delta = Delta(json.loads(item[2]))
        if (
            first_stream_position
            <= current_delta.stream_position
            <= last_stream_position
        ):
            result.append(current_delta)

        if not include_manifest:
            current_delta.manifest = None

    result.sort(reverse=(not ascending_order), key=lambda d: d.stream_position)
    return ListResult.of(result, None, None)


def get_delta(
    namespace: str,
    table_name: str,
    stream_position: int,
    partition_values: Optional[PartitionValues] = None,
    table_version: Optional[str] = None,
    include_manifest: bool = False,
    partition_filter: Optional[PartitionFilter] = None,
    *args,
    **kwargs,
) -> Optional[Delta]:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    stream = get_stream(namespace, table_name, table_version, *args, **kwargs)

    if partition_values is not None and partition_filter is not None:
        raise ValueError(
            "Only one of partition_values or partition_filter must be provided"
        )

    if partition_filter is not None:
        partition_values = partition_filter.partition_values

    partition = get_partition(stream.locator, partition_values, *args, **kwargs)
    delta_locator = DeltaLocator.of(partition.locator, stream_position)

    res = cur.execute(
        "SELECT * FROM deltas WHERE locator = ?", (delta_locator.canonical_string(),)
    )

    serialized_delta = res.fetchone()
    if serialized_delta is None:
        return None

    delta = Delta(json.loads(serialized_delta[2]))

    if not include_manifest:
        delta.manifest = None

    return delta


def get_latest_delta(
    namespace: str,
    table_name: str,
    partition_values: Optional[PartitionValues] = None,
    table_version: Optional[str] = None,
    include_manifest: bool = False,
    partition_filter: Optional[PartitionFilter] = None,
    *args,
    **kwargs,
) -> Optional[Delta]:

    deltas = list_deltas(
        namespace=namespace,
        table_name=table_name,
        partition_values=partition_values,
        table_version=table_version,
        first_stream_position=None,
        last_stream_position=None,
        ascending_order=False,
        include_manifest=include_manifest,
        partition_filter=partition_filter,
        *args,
        **kwargs,
    ).all_items()

    if not deltas:
        return None

    return deltas[0]


def download_delta(
    delta_like: Union[Delta, DeltaLocator],
    table_type: TableType = TableType.PYARROW,
    storage_type: StorageType = StorageType.DISTRIBUTED,
    max_parallelism: Optional[int] = None,
    columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    ray_options_provider: Callable[[int, Any], Dict[str, Any]] = None,
    distributed_dataset_type: DistributedDatasetType = DistributedDatasetType.RAY_DATASET,
    partition_filter: Optional[PartitionFilter] = None,
    *args,
    **kwargs,
) -> Union[LocalDataset, DistributedDataset]:  # type: ignore
    result = []
    if isinstance(delta_like, Delta) and delta_like.manifest is not None:
        manifest = Delta(delta_like).manifest
    else:
        manifest = get_delta_manifest(delta_like, *args, **kwargs)
    partition_values: PartitionValues = None
    if partition_filter is not None:
        partition_values = partition_filter.partition_values
    for entry_index in range(len(manifest.entries)):
        if (
            partition_values is not None
            and partition_values != manifest.entries[entry_index].meta.partition_values
        ):
            continue

        result.append(
            download_delta_manifest_entry(
                delta_like=delta_like,
                entry_index=entry_index,
                table_type=table_type,
                columns=columns,
                file_reader_kwargs_provider=file_reader_kwargs_provider,
                *args,
                **kwargs,
            )
        )

    if storage_type == StorageType.DISTRIBUTED:
        if distributed_dataset_type is DistributedDatasetType.DAFT:
            return daft.from_arrow(result)
        elif distributed_dataset_type is DistributedDatasetType.RAY_DATASET:
            return ray.data.from_arrow(result)
        else:
            raise ValueError(f"Dataset type {distributed_dataset_type} not supported!")

    return result


def download_delta_manifest_entry(
    delta_like: Union[Delta, DeltaLocator],
    entry_index: int,
    table_type: TableType = TableType.PYARROW,
    columns: Optional[List[str]] = None,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
    *args,
    **kwargs,
) -> LocalTable:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    if isinstance(delta_like, Delta) and delta_like.manifest is not None:
        manifest = Delta(delta_like).manifest
    else:
        manifest = get_delta_manifest(delta_like, *args, **kwargs)
    if entry_index >= len(manifest.entries):
        raise IndexError(
            f"Manifest entry index {entry_index} does not exist. "
            f"Valid values: [0, {len(manifest.entries)}]"
        )

    entry = manifest.entries[entry_index]

    res = cur.execute("SELECT value FROM data WHERE uri = ?", (entry.uri,))
    serialized_data = res.fetchone()

    if serialized_data is None:
        raise ValueError(
            f"Invalid value of delta locator: {delta_like.canonical_string()}"
        )

    serialized_data = serialized_data[0]
    if entry.meta.content_type == ContentType.PARQUET:
        if table_type == TableType.PYARROW_PARQUET:
            table = pa.parquet.ParquetFile(io.BytesIO(serialized_data))
        else:
            table = pa.parquet.read_table(io.BytesIO(serialized_data), columns=columns)
    elif entry.meta.content_type == ContentType.UNESCAPED_TSV:
        assert (
            table_type != TableType.PYARROW_PARQUET
        ), f"uTSV table cannot be read as {table_type}"
        parse_options = pa.csv.ParseOptions(delimiter="\t")
        convert_options = pa.csv.ConvertOptions(
            null_values=[""], strings_can_be_null=True, include_columns=columns
        )
        table = pa.csv.read_csv(
            io.BytesIO(serialized_data),
            parse_options=parse_options,
            convert_options=convert_options,
        )
    else:
        raise ValueError(f"Content type: {entry.meta.content_type} not supported.")

    if table_type == TableType.PYARROW:
        return table
    elif table_type == TableType.PYARROW_PARQUET:
        return table
    elif table_type == TableType.NUMPY:
        raise NotImplementedError(f"Table type={table_type} not supported")
    elif table_type == TableType.PANDAS:
        return table.to_pandas()

    return table


def get_delta_manifest(
    delta_like: Union[Delta, DeltaLocator], *args, **kwargs
) -> Optional[Manifest]:
    delta = get_delta(
        namespace=delta_like.namespace,
        table_name=delta_like.table_name,
        stream_position=delta_like.stream_position,
        partition_values=delta_like.partition_values,
        table_version=delta_like.table_version,
        include_manifest=True,
        *args,
        **kwargs,
    )
    if not delta:
        return None

    return delta.manifest


def create_namespace(
    namespace: str, permissions: Dict[str, Any], *args, **kwargs
) -> Namespace:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    locator = NamespaceLocator.of(namespace)
    result = Namespace.of(locator, permissions)
    params = (locator.canonical_string(), json.dumps(result))
    cur.execute(CREATE_NAMESPACES_TABLE)
    cur.execute(CREATE_TABLES_TABLE)
    cur.execute(CREATE_TABLE_VERSIONS_TABLE)
    cur.execute(CREATE_STREAMS_TABLE)
    cur.execute(CREATE_PARTITIONS_TABLE)
    cur.execute(CREATE_DELTAS_TABLE)
    cur.execute(CREATE_DATA_TABLE)
    cur.execute("INSERT OR IGNORE INTO namespaces VALUES(?, ?)", params)
    con.commit()
    return result


def update_namespace(
    namespace: str,
    permissions: Optional[Dict[str, Any]] = None,
    new_namespace: Optional[str] = None,
    *args,
    **kwargs,
) -> None:
    assert new_namespace is None, "namespace name cannot be changed"
    cur, con = _get_sqlite3_cursor_con(kwargs)
    locator = NamespaceLocator.of(namespace)
    result = Namespace.of(locator, permissions)
    params = (json.dumps(result), locator.canonical_string())
    cur.execute("UPDATE namespaces SET value = ? WHERE locator = ?", params)
    con.commit()


def create_table_version(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    schema: Optional[Union[pa.Schema, str, bytes]] = None,
    schema_consistency: Optional[Dict[str, SchemaConsistencyType]] = None,
    partition_keys: Optional[List[Dict[str, Any]]] = None,
    primary_key_column_names: Optional[Set[str]] = None,
    sort_keys: Optional[List[SortKey]] = None,
    table_version_description: Optional[str] = None,
    table_version_properties: Optional[Dict[str, str]] = None,
    table_permissions: Optional[Dict[str, Any]] = None,
    table_description: Optional[str] = None,
    table_properties: Optional[Dict[str, str]] = None,
    supported_content_types: Optional[List[ContentType]] = None,
    partition_spec: Optional[StreamPartitionSpec] = None,
    *args,
    **kwargs,
) -> Stream:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    if partition_keys is not None and partition_spec is not None:
        raise ValueError(
            "Only one of partition_keys or partition_spec must be provided"
        )
    if partition_spec is not None:
        assert (
            partition_spec.ordered_transforms is not None
        ), "Ordered transforms must be specified when partition_spec is specified"
        partition_keys = []
        for transform in partition_spec.ordered_transforms:
            assert transform.name == TransformName.IDENTITY, (
                "Local DeltaCAT storage does not support creating table versions "
                "with non identity transform partition spec"
            )
            transform_params: IdentityTransformParameters = transform.parameters
            partition_keys.append(transform_params.column_name)

    latest_version = get_latest_table_version(namespace, table_name, *args, **kwargs)
    if (
        table_version is not None
        and latest_version
        and int(latest_version.table_version) + 1 != int(table_version)
    ):
        raise AssertionError(
            f"Table version can only be incremented. Last version={latest_version.table_version}"
        )
    elif table_version is None:
        table_version = (
            (int(latest_version.table_version) + 1) if latest_version else "1"
        )

    table_locator = TableLocator.of(NamespaceLocator.of(namespace), table_name)
    table_obj = Table.of(
        table_locator, table_permissions, table_description, table_properties
    )
    table_version_locator = TableVersionLocator.of(
        table_locator=table_locator, table_version=table_version
    )

    stream_id = uuid.uuid4().__str__()

    if table_version_properties is None:
        table_version_properties = {}

    properties = {**table_version_properties, STREAM_ID_PROPERTY: stream_id}
    table_version_obj = TableVersion.of(
        table_version_locator,
        schema=schema,
        partition_keys=partition_keys,
        primary_key_columns=primary_key_column_names,
        description=table_version_description,
        properties=properties,
        sort_keys=sort_keys,
        content_types=supported_content_types,
    )
    stream_locator = StreamLocator.of(
        table_version_obj.locator, stream_id=stream_id, storage_type=STORAGE_TYPE
    )
    result_stream = Stream.of(
        stream_locator, partition_keys=partition_keys, state=CommitState.COMMITTED
    )

    params = (
        table_locator.canonical_string(),
        table_locator.namespace_locator.canonical_string(),
        json.dumps(table_obj),
    )
    cur.execute("INSERT OR IGNORE INTO tables VALUES (?, ?, ?)", params)
    params = (
        table_version_locator.canonical_string(),
        table_locator.canonical_string(),
        json.dumps(table_version_obj),
    )
    cur.execute("INSERT OR IGNORE INTO table_versions VALUES (?, ?, ?)", params)

    params = (
        stream_locator.canonical_string(),
        table_version_locator.canonical_string(),
        json.dumps(result_stream),
    )
    cur.execute("INSERT OR IGNORE INTO streams VALUES (?, ?, ?)", params)
    con.commit()
    return result_stream


def update_table(
    namespace: str,
    table_name: str,
    permissions: Optional[Dict[str, Any]] = None,
    description: Optional[str] = None,
    properties: Optional[Dict[str, str]] = None,
    new_table_name: Optional[str] = None,
    *args,
    **kwargs,
) -> None:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    table_locator = TableLocator.of(NamespaceLocator.of(namespace), table_name)
    table_obj = Table.of(table_locator, permissions, description, properties)

    params = (table_locator.canonical_string(),)
    cur.execute("DELETE FROM tables WHERE locator = ?", params)
    params = (
        table_locator.canonical_string(),
        table_locator.namespace_locator.canonical_string(),
        json.dumps(table_obj),
    )
    cur.execute("INSERT INTO tables VALUES (?, ?, ?)", params)
    con.commit()


def update_table_version(
    namespace: str,
    table_name: str,
    table_version: str,
    lifecycle_state: Optional[LifecycleState] = None,
    schema: Optional[Union[pa.Schema, str, bytes]] = None,
    schema_consistency: Optional[Dict[str, SchemaConsistencyType]] = None,
    description: Optional[str] = None,
    properties: Optional[Dict[str, str]] = None,
    *args,
    **kwargs,
) -> None:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    table_locator = TableLocator.of(NamespaceLocator.of(namespace), table_name)
    table_version_locator = TableVersionLocator.of(
        table_locator=table_locator, table_version=table_version
    )

    res = cur.execute(
        "SELECT * from table_versions WHERE locator = ?",
        (table_version_locator.canonical_string(),),
    )
    serialized_table_version = res.fetchone()
    assert (
        serialized_table_version is not None
    ), f"Table version not found with locator={table_version_locator.canonical_string()}"
    current_table_version_obj = TableVersion(json.loads(serialized_table_version[2]))

    if properties is None:
        properties = {}

    current_props = (
        current_table_version_obj.properties
        if current_table_version_obj.properties
        else {}
    )

    tv_properties = {**properties, **current_props}
    table_version_obj = TableVersion.of(
        table_version_locator,
        schema=schema,
        partition_keys=current_table_version_obj.partition_keys,
        primary_key_columns=current_table_version_obj.primary_keys,
        description=description,
        properties=tv_properties,
        sort_keys=current_table_version_obj.sort_keys,
        content_types=current_table_version_obj.content_types,
    )

    params = (
        table_locator.canonical_string(),
        json.dumps(table_version_obj),
        table_version_locator.canonical_string(),
    )
    cur.execute(
        "UPDATE table_versions SET table_locator = ?, value = ? WHERE locator = ?",
        params,
    )
    con.commit()


def stage_stream(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs,
) -> Stream:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    existing_table_version = get_table_version(
        namespace, table_name, table_version, *args, **kwargs
    )
    existing_stream = get_stream(namespace, table_name, table_version, *args, **kwargs)

    stream_id = uuid.uuid4().__str__()
    new_stream_locator = StreamLocator.of(
        existing_table_version.locator, stream_id, STORAGE_TYPE
    )
    new_stream = Stream.of(
        new_stream_locator,
        existing_stream.partition_keys,
        CommitState.STAGED,
        existing_stream.locator.canonical_string(),
    )

    params = (
        new_stream_locator.canonical_string(),
        existing_table_version.locator.canonical_string(),
        json.dumps(new_stream),
    )
    cur.execute("INSERT INTO streams VALUES (?, ?, ?)", params)
    con.commit()

    return new_stream


def commit_stream(stream: Stream, *args, **kwargs) -> Stream:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    existing_table_version = get_table_version(
        stream.namespace, stream.table_name, stream.table_version, *args, **kwargs
    )
    stream_to_commit = Stream.of(
        stream.locator,
        stream.partition_keys,
        CommitState.COMMITTED,
        stream.previous_stream_digest,
    )

    existing_table_version.properties[
        STREAM_ID_PROPERTY
    ] = stream_to_commit.locator.stream_id

    params = (
        json.dumps(existing_table_version),
        existing_table_version.locator.canonical_string(),
    )
    cur.execute("UPDATE table_versions SET value = ? WHERE locator = ?", params)
    params = (json.dumps(stream_to_commit), stream_to_commit.locator.canonical_string())
    cur.execute("UPDATE streams SET value = ? WHERE locator = ?", params)
    con.commit()

    return stream_to_commit


def delete_stream(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs,
) -> None:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    table_version_locator = TableVersionLocator.of(
        TableLocator.of(NamespaceLocator.of(namespace), table_name), table_version
    )

    res = cur.execute(
        "SELECT locator FROM streams WHERE table_version_locator = ?",
        (table_version_locator.canonical_string(),),
    )
    locators = res.fetchall()
    cur.executemany("DELETE FROM streams WHERE locator = ?", locators)
    cur.execute(
        "DELETE FROM table_versions WHERE locator = ?",
        (table_version_locator.canonical_string(),),
    )

    con.commit()


def stage_partition(
    stream: Stream, partition_values: Optional[PartitionValues] = None, *args, **kwargs
) -> Partition:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    partition_id = uuid.uuid4().__str__()
    partition_locator = PartitionLocator.of(
        stream.locator, partition_values=partition_values, partition_id=partition_id
    )

    tv = get_table_version(
        stream.namespace, stream.table_name, stream.table_version, *args, **kwargs
    )

    pv_partition = get_partition(
        stream.locator, partition_values=partition_values, *args, **kwargs
    )

    stream_position = current_time_ms()
    partition = Partition.of(
        partition_locator,
        schema=tv.schema,
        content_types=tv.content_types,
        state=CommitState.STAGED,
        previous_stream_position=pv_partition.stream_position if pv_partition else None,
        previous_partition_id=pv_partition.partition_id if pv_partition else None,
        stream_position=stream_position,
    )

    params = (
        partition.locator.canonical_string(),
        partition.stream_locator.canonical_string(),
        json.dumps(partition),
    )
    cur.execute("INSERT INTO partitions VALUES (?, ?, ?)", params)
    con.commit()

    return partition


def commit_partition(
    partition: Partition,
    previous_partition: Optional[Partition] = None,
    *args,
    **kwargs,
) -> Partition:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    pv_partition: Optional[Partition] = previous_partition or get_partition(
        partition.stream_locator,
        partition_values=partition.partition_values,
        *args,
        **kwargs,
    )
    # deprecate old partition and commit new one
    if pv_partition:
        pv_partition.state = CommitState.DEPRECATED
        params = (json.dumps(pv_partition), pv_partition.locator.canonical_string())
        cur.execute("UPDATE partitions SET value = ? WHERE locator = ?", params)
    previous_partition_deltas = (
        list_partition_deltas(
            pv_partition, ascending_order=False, *args, **kwargs
        ).all_items()
        or []
    )

    partition_deltas: Optional[List[Delta]] = (
        list_partition_deltas(
            partition, ascending_order=False, *args, **kwargs
        ).all_items()
        or []
    )

    # if previous_partition is passed in, table is in-place compacted and we need to run merge-and-promote
    if previous_partition:
        partition_deltas = _merge_and_promote(
            partition_deltas, previous_partition_deltas
        )

    stream_position = (
        partition_deltas[0].stream_position
        if partition_deltas
        else partition.stream_position
    )

    partition.stream_position = stream_position
    if partition_deltas:
        partition.locator = partition_deltas[0].partition_locator

    partition.state = CommitState.COMMITTED
    partition.previous_stream_position = (
        pv_partition.stream_position if pv_partition else None
    )
    params = (json.dumps(partition), partition.locator.canonical_string())
    cur.execute("UPDATE partitions SET value = ? WHERE locator = ?", params)
    con.commit()

    return partition


def delete_partition(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    partition_values: Optional[PartitionValues] = None,
    *args,
    **kwargs,
) -> None:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    stream = get_stream(namespace, table_name, table_version, *args, **kwargs)
    partition = get_partition(stream.locator, partition_values, *args, **kwargs)

    partition.state = CommitState.DEPRECATED
    params = (json.dumps(partition), partition.locator.canonical_string())

    cur.execute("UPDATE partitions SET value = ? WHERE locator = ?", params)
    con.commit()


def get_partition(
    stream_locator: StreamLocator,
    partition_values: Optional[PartitionValues] = None,
    *args,
    **kwargs,
) -> Optional[Partition]:
    cur, con = _get_sqlite3_cursor_con(kwargs)

    res = cur.execute(
        "SELECT * FROM partitions WHERE stream_locator = ?",
        (stream_locator.canonical_string(),),
    )

    serialized_partitions = res.fetchall()

    if not serialized_partitions:
        return None

    if partition_values is None:
        partition_values = []

    prior_pv = ",".join(partition_values)

    for item in serialized_partitions:
        partition = Partition(json.loads(item[2]))
        pv = ",".join(partition.partition_values if partition.partition_values else [])

        if pv == prior_pv and partition.state == CommitState.COMMITTED:
            return partition

    return None


def stage_delta(
    data: Union[LocalTable, LocalDataset, DistributedDataset, Manifest],
    partition: Partition,
    delta_type: DeltaType = DeltaType.UPSERT,
    max_records_per_entry: Optional[int] = None,
    author: Optional[ManifestAuthor] = None,
    properties: Optional[Dict[str, str]] = None,
    s3_table_writer_kwargs: Optional[Dict[str, Any]] = None,
    content_type: ContentType = ContentType.PARQUET,
    delete_parameters: Optional[DeleteParameters] = None,
    partition_spec: Optional[DeltaPartitionSpec] = None,
    partition_values: Optional[PartitionValues] = None,
    *args,
    **kwargs,
) -> Delta:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    manifest_id = uuid.uuid4().__str__()
    uri = _get_manifest_entry_uri(manifest_id)

    if data is None:
        delta = create_empty_delta(
            partition,
            delta_type,
            author,
            properties=properties,
            manifest_entry_id=manifest_id,
        )
        cur.execute("INSERT OR IGNORE INTO data VALUES (?, ?)", (uri, None))
        params = (delta.locator.canonical_string(), "staged_delta", json.dumps(delta))
        cur.execute("INSERT OR IGNORE INTO deltas VALUES (?, ?, ?)", params)
        con.commit()
        return delta

    if partition_spec:
        assert partition_values is not None, (
            "partition_values must be provided as local "
            "storage does not support computing it from input data"
        )

    serialized_data = None
    if content_type == ContentType.PARQUET:
        buffer = io.BytesIO()
        pa.parquet.write_table(data, buffer)
        serialized_data = buffer.getvalue()
    elif content_type == ContentType.UNESCAPED_TSV:
        buffer = io.BytesIO()
        write_options = pa.csv.WriteOptions(
            include_header=True, delimiter="\t", quoting_style="none"
        )
        pa.csv.write_csv(data, buffer, write_options=write_options)
        serialized_data = buffer.getvalue()
    else:
        raise ValueError(f"Unsupported content type: {content_type}")

    stream_position = current_time_ms()
    delta_locator = DeltaLocator.of(partition.locator, stream_position=stream_position)

    meta = ManifestMeta.of(
        len(data),
        len(serialized_data),
        content_type=content_type,
        content_encoding=ContentEncoding.IDENTITY,
        source_content_length=data.nbytes,
        partition_values=partition_values,
    )

    manifest = Manifest.of(
        entries=ManifestEntryList.of(
            [
                ManifestEntry.of(
                    uri=uri, url=uri, meta=meta, mandatory=True, uuid=manifest_id
                )
            ]
        ),
        author=author,
        uuid=manifest_id,
    )

    delta = Delta.of(
        delta_locator,
        delta_type=delta_type,
        meta=meta,
        properties=properties,
        manifest=manifest,
        previous_stream_position=partition.stream_position,
        delete_parameters=delete_parameters,
    )

    params = (uri, serialized_data)
    cur.execute("INSERT OR IGNORE INTO data VALUES (?, ?)", params)

    params = (delta_locator.canonical_string(), "staged_delta", json.dumps(delta))
    cur.execute("INSERT OR IGNORE INTO deltas VALUES (?, ?, ?)", params)

    con.commit()
    return delta


def commit_delta(delta: Delta, *args, **kwargs) -> Delta:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    delta_stream_position: Optional[int] = delta.stream_position
    delta.locator.stream_position = delta_stream_position or current_time_ms()

    params = (
        delta.locator.canonical_string(),
        delta.partition_locator.canonical_string(),
        json.dumps(delta),
    )

    cur.execute("INSERT OR IGNORE INTO deltas VALUES (?, ?, ?)", params)

    params = (
        delta.partition_locator.canonical_string(),
        json.dumps(delta),
        delta.locator.canonical_string(),
    )
    cur.execute(
        "UPDATE deltas SET partition_locator = ?, value = ? WHERE locator = ?", params
    )
    con.commit()
    return delta


def get_namespace(namespace: str, *args, **kwargs) -> Optional[Namespace]:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    locator = NamespaceLocator.of(namespace)

    res = cur.execute(
        "SELECT * FROM namespaces WHERE locator = ?", (locator.canonical_string(),)
    )
    serialized_result = res.fetchone()

    if serialized_result is None:
        return None

    return Namespace(json.loads(serialized_result[1]))


def namespace_exists(namespace: str, *args, **kwargs) -> bool:
    obj = get_namespace(namespace, *args, **kwargs)

    return obj is not None


def get_table(namespace: str, table_name: str, *args, **kwargs) -> Optional[Table]:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    locator = TableLocator.of(NamespaceLocator.of(namespace), table_name)

    res = cur.execute(
        "SELECT * FROM tables WHERE locator = ?", (locator.canonical_string(),)
    )
    serialized_result = res.fetchone()

    if serialized_result is None:
        return None

    return Table(json.loads(serialized_result[2]))


def table_exists(namespace: str, table_name: str, *args, **kwargs) -> bool:
    obj = get_table(namespace, table_name, *args, **kwargs)

    return obj is not None


def get_table_version(
    namespace: str, table_name: str, table_version: str, *args, **kwargs
) -> Optional[TableVersion]:
    cur, con = _get_sqlite3_cursor_con(kwargs)
    locator = TableVersionLocator.of(
        TableLocator.of(NamespaceLocator.of(namespace), table_name), table_version
    )

    res = cur.execute(
        "SELECT * FROM table_versions WHERE locator = ?", (locator.canonical_string(),)
    )
    serialized_table_version = res.fetchone()

    if serialized_table_version is None:
        return None

    return TableVersion(json.loads(serialized_table_version[2]))


def get_latest_table_version(
    namespace: str, table_name: str, *args, **kwargs
) -> Optional[TableVersion]:
    table_versions = list_table_versions(
        namespace, table_name, *args, **kwargs
    ).all_items()
    if not table_versions:
        return None

    table_versions.sort(reverse=True, key=lambda v: int(v.table_version))
    return table_versions[0]


def get_latest_active_table_version(
    namespace: str, table_name: str, *args, **kwargs
) -> Optional[TableVersion]:

    # This module does not support table version lifecycle state
    return get_latest_table_version(namespace, table_name, *args, **kwargs)


def get_table_version_schema(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs,
) -> Optional[Union[pa.Schema, str, bytes]]:
    obj = get_table_version(namespace, table_name, table_version, *args, **kwargs)

    return obj.schema


def table_version_exists(
    namespace: str, table_name: str, table_version: str, *args, **kwargs
) -> bool:
    obj = get_table_version(namespace, table_name, table_version, *args, **kwargs)

    return obj is not None


def get_stream(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs,
) -> Optional[Stream]:
    assert not isinstance(table_version, int), f"Passed an integer as the table version"
    obj = get_table_version(namespace, table_name, table_version, *args, **kwargs)

    if obj is None:
        return None

    stream_id = obj.properties.get(STREAM_ID_PROPERTY)
    if stream_id is None:
        return None

    cur, con = _get_sqlite3_cursor_con(kwargs)
    stream_locator = StreamLocator.of(
        obj.locator, stream_id=stream_id, storage_type=STORAGE_TYPE
    )
    res = cur.execute(
        "SELECT * FROM streams WHERE locator = ?", (stream_locator.canonical_string(),)
    )

    serialized_stream = res.fetchone()
    if serialized_stream is None:
        return None

    return Stream(json.loads(serialized_stream[2]))


def get_table_version_column_names(
    namespace: str,
    table_name: str,
    table_version: Optional[str] = None,
    *args,
    **kwargs,
) -> Optional[List[str]]:
    raise NotImplementedError("Fetching column names is not supported")


def can_categorize(e: BaseException, **kwargs) -> bool:
    if isinstance(e, InvalidNamespaceError):
        return True
    else:
        return False


def raise_categorized_error(e: BaseException, **kwargs):
    if isinstance(e, InvalidNamespaceError):
        raise LocalStorageValidationError("Namespace provided is invalid!")



from typing import Dict, Any
import ray
import os
import pytest
import boto3
from deltacat.compute.compactor.model.compaction_session_audit_info import (
    CompactionSessionAuditInfo,
)
from boto3.resources.base import ServiceResource
import deltacat.tests.local_deltacat_storage as ds
from deltacat.types.media import ContentType
from deltacat.compute.compactor_v2.compaction_session import (
    compact_partition,
)
from deltacat.compute.compactor.model.compact_partition_params import (
    CompactPartitionParams,
)
from deltacat.tests.test_utils.utils import read_s3_contents
from deltacat.tests.compute.test_util_constant import (
    TEST_S3_RCF_BUCKET_NAME,
)
from deltacat.compute.resource_estimation import ResourceEstimationMethod
from deltacat.tests.compute.test_util_common import get_rcf
from deltacat.tests.test_utils.pyarrow import (
    stage_partition_from_file_paths,
    commit_delta_to_staged_partition,
    commit_delta_to_partition,
)
from moto import mock_s3

DATABASE_FILE_PATH_KEY, DATABASE_FILE_PATH_VALUE = (
    "db_file_path",
    "deltacat/tests/local_deltacat_storage/db_test.sqlite",
)


@pytest.fixture(autouse=True, scope="module")
def setup_ray_cluster():
    ray.init(local_mode=True, ignore_reinit_error=True)
    yield
    ray.shutdown()


@pytest.fixture(autouse=True, scope="module")
def mock_aws_credential():
    os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    os.environ["AWS_SECRET_ACCESS_ID"] = "testing"
    os.environ["AWS_SECURITY_TOKEN"] = "testing"
    os.environ["AWS_SESSION_TOKEN"] = "testing"
    os.environ["AWS_DEFAULT_REGION"] = "us-east-1"
    yield


@pytest.fixture(scope="module")
def s3_resource(mock_aws_credential):
    with mock_s3():
        yield boto3.resource("s3")


@pytest.fixture(autouse=True, scope="module")
def setup_compaction_artifacts_s3_bucket(s3_resource: ServiceResource):
    s3_resource.create_bucket(
        ACL="authenticated-read",
        Bucket=TEST_S3_RCF_BUCKET_NAME,
    )
    yield


@pytest.fixture(scope="function")
def local_deltacat_storage_kwargs(request: pytest.FixtureRequest):
    kwargs_for_local_deltacat_storage: Dict[str, Any] = {
        DATABASE_FILE_PATH_KEY: DATABASE_FILE_PATH_VALUE,
    }
    yield kwargs_for_local_deltacat_storage
    if os.path.exists(DATABASE_FILE_PATH_VALUE):
        os.remove(DATABASE_FILE_PATH_VALUE)


class TestCompactionSession:
    """
    This class adds specific tests that aren't part of the parametrized test suite.
    """

    NAMESPACE = "compact_partition_v2_namespace"
    BACKFILL_FILE_PATH = (
        "deltacat/tests/compute/compactor_v2/data/backfill_source_date_pk.csv"
    )
    INCREMENTAL_FILE_PATH = (
        "deltacat/tests/compute/compactor_v2/data/incremental_source_date_pk.csv"
    )
    ERROR_RATE = 0.05

    def test_compact_partition_when_no_input_deltas_to_compact(
        self, local_deltacat_storage_kwargs
    ):
        # setup
        staged_source = stage_partition_from_file_paths(
            self.NAMESPACE, ["test"], **local_deltacat_storage_kwargs
        )
        source_partition = ds.commit_partition(
            staged_source, **local_deltacat_storage_kwargs
        )

        staged_dest = stage_partition_from_file_paths(
            self.NAMESPACE, ["destination"], **local_deltacat_storage_kwargs
        )
        dest_partition = ds.commit_partition(
            staged_dest, **local_deltacat_storage_kwargs
        )

        # action
        rcf_url = compact_partition(
            CompactPartitionParams.of(
                {
                    "compaction_artifact_s3_bucket": TEST_S3_RCF_BUCKET_NAME,
                    "compacted_file_content_type": ContentType.PARQUET,
                    "dd_max_parallelism_ratio": 1.0,
                    "deltacat_storage": ds,
                    "deltacat_storage_kwargs": local_deltacat_storage_kwargs,
                    "destination_partition_locator": dest_partition.locator,
                    "drop_duplicates": True,
                    "hash_bucket_count": 2,
                    "last_stream_position_to_compact": source_partition.stream_position,
                    "list_deltas_kwargs": {
                        **local_deltacat_storage_kwargs,
                        **{"equivalent_table_types": []},
                    },
                    "primary_keys": ["pk"],
                    "rebase_source_partition_locator": None,
                    "rebase_source_partition_high_watermark": None,
                    "records_per_compacted_file": 4000,
                    "s3_client_kwargs": {},
                    "source_partition_locator": source_partition.locator,
                }
            )
        )

        # verify that no RCF is written
        assert rcf_url is None

    def test_compact_partition_when_rcf_was_written_by_past_commit(
        self, s3_resource, local_deltacat_storage_kwargs
    ):
        """
        Backward compatibility test for when a RCF was written by a previous commit.
        """

        # setup
        staged_source = stage_partition_from_file_paths(
            self.NAMESPACE, ["source"], **local_deltacat_storage_kwargs
        )

        source_delta = commit_delta_to_staged_partition(
            staged_source, [self.BACKFILL_FILE_PATH], **local_deltacat_storage_kwargs
        )

        staged_dest = stage_partition_from_file_paths(
            self.NAMESPACE, ["destination"], **local_deltacat_storage_kwargs
        )
        dest_partition = ds.commit_partition(
            staged_dest, **local_deltacat_storage_kwargs
        )

        # action
        rcf_url = compact_partition(
            CompactPartitionParams.of(
                {
                    "compaction_artifact_s3_bucket": TEST_S3_RCF_BUCKET_NAME,
                    "compacted_file_content_type": ContentType.PARQUET,
                    "dd_max_parallelism_ratio": 1.0,
                    "deltacat_storage": ds,
                    "deltacat_storage_kwargs": local_deltacat_storage_kwargs,
                    "destination_partition_locator": dest_partition.locator,
                    "drop_duplicates": True,
                    "hash_bucket_count": 1,
                    "last_stream_position_to_compact": source_delta.stream_position,
                    "list_deltas_kwargs": {
                        **local_deltacat_storage_kwargs,
                        **{"equivalent_table_types": []},
                    },
                    "primary_keys": [],
                    "rebase_source_partition_locator": source_delta.partition_locator,
                    "rebase_source_partition_high_watermark": source_delta.stream_position,
                    "records_per_compacted_file": 4000,
                    "s3_client_kwargs": {},
                    "source_partition_locator": source_delta.partition_locator,
                }
            )
        )

        bucket, backfill_key1, backfill_key2 = rcf_url.strip("s3://").split("/")
        assert bucket == TEST_S3_RCF_BUCKET_NAME

        # Now delete the RCF at new location and copy it to old location
        # Copy the RCF from rcf_url to another location
        s3_resource.Object(TEST_S3_RCF_BUCKET_NAME, f"{backfill_key1}.json").copy_from(
            CopySource=f"{TEST_S3_RCF_BUCKET_NAME}/{backfill_key1}/{backfill_key2}"
        )

        s3_resource.Object(
            TEST_S3_RCF_BUCKET_NAME, f"{backfill_key1}/{backfill_key2}"
        ).delete()

        # Now run an incremental compaction and verify if the previous RCF was read properly.

        new_source_delta = commit_delta_to_partition(
            source_delta.partition_locator,
            [self.INCREMENTAL_FILE_PATH],
            **local_deltacat_storage_kwargs,
        )

        new_rcf_url = compact_partition(
            CompactPartitionParams.of(
                {
                    "compaction_artifact_s3_bucket": TEST_S3_RCF_BUCKET_NAME,
                    "compacted_file_content_type": ContentType.PARQUET,
                    "dd_max_parallelism_ratio": 1.0,
                    "deltacat_storage": ds,
                    "deltacat_storage_kwargs": local_deltacat_storage_kwargs,
                    "destination_partition_locator": dest_partition.locator,
                    "drop_duplicates": True,
                    "hash_bucket_count": 1,
                    "last_stream_position_to_compact": new_source_delta.stream_position,
                    "list_deltas_kwargs": {
                        **local_deltacat_storage_kwargs,
                        **{"equivalent_table_types": []},
                    },
                    "primary_keys": ["pk"],
                    "rebase_source_partition_locator": None,
                    "rebase_source_partition_high_watermark": None,
                    "records_per_compacted_file": 4000,
                    "s3_client_kwargs": {},
                    "source_partition_locator": new_source_delta.partition_locator,
                }
            )
        )

        new_bucket, incremental_key1, incremental_key2 = new_rcf_url.strip(
            "s3://"
        ).split("/")

        assert new_bucket == TEST_S3_RCF_BUCKET_NAME
        assert backfill_key1 == incremental_key1
        assert backfill_key2 != incremental_key2

        rcf = get_rcf(s3_resource, new_rcf_url)

        _, compaction_audit_key = rcf.compaction_audit_url.strip("s3://").split("/", 1)
        compaction_audit = CompactionSessionAuditInfo(
            **read_s3_contents(
                s3_resource, TEST_S3_RCF_BUCKET_NAME, compaction_audit_key
            )
        )

        # as it should be running incremental
        assert compaction_audit.uniform_deltas_created == 1
        assert compaction_audit.input_records == 6

    def test_compact_partition_when_incremental_then_rcf_stats_accurate(
        self, s3_resource, local_deltacat_storage_kwargs
    ):
        """
        A test case which asserts the RCF stats are correctly generated for
        a rebase and incremental use-case.
        """

        # setup
        staged_source = stage_partition_from_file_paths(
            self.NAMESPACE, ["source"], **local_deltacat_storage_kwargs
        )

        source_delta = commit_delta_to_staged_partition(
            staged_source, [self.BACKFILL_FILE_PATH], **local_deltacat_storage_kwargs
        )

        staged_dest = stage_partition_from_file_paths(
            self.NAMESPACE, ["destination"], **local_deltacat_storage_kwargs
        )
        dest_partition = ds.commit_partition(
            staged_dest, **local_deltacat_storage_kwargs
        )

        # action
        rcf_url = compact_partition(
            CompactPartitionParams.of(
                {
                    "compaction_artifact_s3_bucket": TEST_S3_RCF_BUCKET_NAME,
                    "compacted_file_content_type": ContentType.PARQUET,
                    "dd_max_parallelism_ratio": 1.0,
                    "deltacat_storage": ds,
                    "deltacat_storage_kwargs": local_deltacat_storage_kwargs,
                    "destination_partition_locator": dest_partition.locator,
                    "drop_duplicates": True,
                    "hash_bucket_count": 2,
                    "last_stream_position_to_compact": source_delta.stream_position,
                    "list_deltas_kwargs": {
                        **local_deltacat_storage_kwargs,
                        **{"equivalent_table_types": []},
                    },
                    "primary_keys": ["pk"],
                    "rebase_source_partition_locator": source_delta.partition_locator,
                    "rebase_source_partition_high_watermark": source_delta.stream_position,
                    "records_per_compacted_file": 4000,
                    "s3_client_kwargs": {},
                    "source_partition_locator": source_delta.partition_locator,
                }
            )
        )

        backfill_rcf = get_rcf(s3_resource, rcf_url)
        _, compaction_audit_key = backfill_rcf.compaction_audit_url.strip(
            "s3://"
        ).split("/", 1)
        compaction_audit = CompactionSessionAuditInfo(
            **read_s3_contents(
                s3_resource, TEST_S3_RCF_BUCKET_NAME, compaction_audit_key
            )
        )

        assert abs(backfill_rcf.input_inflation - 0.05235042735042735) <= 1e-5
        assert abs(backfill_rcf.input_average_record_size_bytes - 12.25) <= 1e-5

        assert compaction_audit.input_records == 4
        assert compaction_audit.records_deduped == 0
        assert compaction_audit.records_deleted == 0
        assert compaction_audit.untouched_file_count == 0
        assert compaction_audit.untouched_record_count == 0
        assert compaction_audit.untouched_size_bytes == 0
        assert compaction_audit.untouched_file_ratio == 0
        assert compaction_audit.uniform_deltas_created == 1
        assert compaction_audit.hash_bucket_count == 2
        assert compaction_audit.input_file_count == 1
        assert compaction_audit.output_file_count == 2
        assert abs(compaction_audit.output_size_bytes - 1832) / 1832 <= self.ERROR_RATE
        assert abs(compaction_audit.input_size_bytes - 936) / 936 <= self.ERROR_RATE

        # Now run an incremental compaction and verify if the previous RCF was read properly.
        new_source_delta = commit_delta_to_partition(
            source_delta.partition_locator,
            [self.INCREMENTAL_FILE_PATH],
            **local_deltacat_storage_kwargs,
        )

        new_destination_partition = ds.get_partition(
            dest_partition.stream_locator, [], **local_deltacat_storage_kwargs
        )

        new_rcf_url = compact_partition(
            CompactPartitionParams.of(
                {
                    "compaction_artifact_s3_bucket": TEST_S3_RCF_BUCKET_NAME,
                    "compacted_file_content_type": ContentType.PARQUET,
                    "dd_max_parallelism_ratio": 1.0,
                    "deltacat_storage": ds,
                    "deltacat_storage_kwargs": local_deltacat_storage_kwargs,
                    "destination_partition_locator": new_destination_partition.locator,
                    "drop_duplicates": True,
                    "hash_bucket_count": 2,
                    "last_stream_position_to_compact": new_source_delta.stream_position,
                    "list_deltas_kwargs": {
                        **local_deltacat_storage_kwargs,
                        **{"equivalent_table_types": []},
                    },
                    "primary_keys": ["pk"],
                    "rebase_source_partition_locator": None,
                    "rebase_source_partition_high_watermark": None,
                    "records_per_compacted_file": 4000,
                    "s3_client_kwargs": {},
                    "source_partition_locator": new_source_delta.partition_locator,
                }
            )
        )

        new_rcf = get_rcf(s3_resource, new_rcf_url)
        _, compaction_audit_key = new_rcf.compaction_audit_url.strip("s3://").split(
            "/", 1
        )
        compaction_audit = CompactionSessionAuditInfo(
            **read_s3_contents(
                s3_resource, TEST_S3_RCF_BUCKET_NAME, compaction_audit_key
            )
        )

        # as it should be running incremental
        assert abs(new_rcf.input_inflation - 0.027292576419213975) <= 1e-5
        assert abs(new_rcf.input_average_record_size_bytes - 12.5) <= 1e-5

        assert compaction_audit.input_records == 6
        assert compaction_audit.records_deduped == 1
        assert compaction_audit.records_deleted == 0
        assert compaction_audit.untouched_file_count == 1
        assert compaction_audit.untouched_record_count == 2
        assert (
            abs(compaction_audit.untouched_size_bytes - 916) / 916 <= self.ERROR_RATE
        )  # 5% error
        assert abs(compaction_audit.untouched_file_ratio - 50) <= 1e-5
        assert compaction_audit.uniform_deltas_created == 1
        assert compaction_audit.hash_bucket_count == 2
        assert compaction_audit.input_file_count == 3
        assert compaction_audit.output_file_count == 2
        assert abs(compaction_audit.output_size_bytes - 1843) / 1843 <= self.ERROR_RATE
        assert abs(compaction_audit.input_size_bytes - 2748) / 2748 <= self.ERROR_RATE

    def test_compact_partition_when_incremental_then_intelligent_estimation_sanity(
        self, s3_resource, local_deltacat_storage_kwargs
    ):
        """
        A test case which asserts the RCF stats are correctly generated for
        a rebase and incremental use-case.
        """

        # setup
        staged_source = stage_partition_from_file_paths(
            self.NAMESPACE, ["source"], **local_deltacat_storage_kwargs
        )

        source_delta = commit_delta_to_staged_partition(
            staged_source, [self.BACKFILL_FILE_PATH], **local_deltacat_storage_kwargs
        )

        staged_dest = stage_partition_from_file_paths(
            self.NAMESPACE, ["destination"], **local_deltacat_storage_kwargs
        )
        dest_partition = ds.commit_partition(
            staged_dest, **local_deltacat_storage_kwargs
        )

        # action
        compact_partition(
            CompactPartitionParams.of(
                {
                    "compaction_artifact_s3_bucket": TEST_S3_RCF_BUCKET_NAME,
                    "compacted_file_content_type": ContentType.PARQUET,
                    "dd_max_parallelism_ratio": 1.0,
                    "deltacat_storage": ds,
                    "deltacat_storage_kwargs": local_deltacat_storage_kwargs,
                    "destination_partition_locator": dest_partition.locator,
                    "drop_duplicates": True,
                    "hash_bucket_count": 2,
                    "last_stream_position_to_compact": source_delta.stream_position,
                    "list_deltas_kwargs": {
                        **local_deltacat_storage_kwargs,
                        **{"equivalent_table_types": []},
                    },
                    "primary_keys": ["pk"],
                    "rebase_source_partition_locator": source_delta.partition_locator,
                    "rebase_source_partition_high_watermark": source_delta.stream_position,
                    "records_per_compacted_file": 4000,
                    "s3_client_kwargs": {},
                    "source_partition_locator": source_delta.partition_locator,
                    "resource_estimation_method": ResourceEstimationMethod.INTELLIGENT_ESTIMATION,
                }
            )
        )

    def test_compact_partition_when_incremental_then_content_type_meta_estimation_sanity(
        self, s3_resource, local_deltacat_storage_kwargs
    ):
        """
        A test case which asserts the RCF stats are correctly generated for
        a rebase and incremental use-case.
        """

        # setup
        staged_source = stage_partition_from_file_paths(
            self.NAMESPACE, ["source"], **local_deltacat_storage_kwargs
        )

        source_delta = commit_delta_to_staged_partition(
            staged_source, [self.BACKFILL_FILE_PATH], **local_deltacat_storage_kwargs
        )

        staged_dest = stage_partition_from_file_paths(
            self.NAMESPACE, ["destination"], **local_deltacat_storage_kwargs
        )
        dest_partition = ds.commit_partition(
            staged_dest, **local_deltacat_storage_kwargs
        )

        # action
        compact_partition(
            CompactPartitionParams.of(
                {
                    "compaction_artifact_s3_bucket": TEST_S3_RCF_BUCKET_NAME,
                    "compacted_file_content_type": ContentType.PARQUET,
                    "dd_max_parallelism_ratio": 1.0,
                    "deltacat_storage": ds,
                    "deltacat_storage_kwargs": local_deltacat_storage_kwargs,
                    "destination_partition_locator": dest_partition.locator,
                    "drop_duplicates": True,
                    "hash_bucket_count": 2,
                    "last_stream_position_to_compact": source_delta.stream_position,
                    "list_deltas_kwargs": {
                        **local_deltacat_storage_kwargs,
                        **{"equivalent_table_types": []},
                    },
                    "primary_keys": ["pk"],
                    "rebase_source_partition_locator": source_delta.partition_locator,
                    "rebase_source_partition_high_watermark": source_delta.stream_position,
                    "records_per_compacted_file": 4000,
                    "s3_client_kwargs": {},
                    "source_partition_locator": source_delta.partition_locator,
                    "resource_estimation_method": ResourceEstimationMethod.CONTENT_TYPE_META,
                }
            )
        )

    def test_compact_partition_when_incremental_then_previous_inflation_estimation_sanity(
        self, s3_resource, local_deltacat_storage_kwargs
    ):
        """
        A test case which asserts the RCF stats are correctly generated for
        a rebase and incremental use-case.
        """

        # setup
        staged_source = stage_partition_from_file_paths(
            self.NAMESPACE, ["source"], **local_deltacat_storage_kwargs
        )

        source_delta = commit_delta_to_staged_partition(
            staged_source, [self.BACKFILL_FILE_PATH], **local_deltacat_storage_kwargs
        )

        staged_dest = stage_partition_from_file_paths(
            self.NAMESPACE, ["destination"], **local_deltacat_storage_kwargs
        )
        dest_partition = ds.commit_partition(
            staged_dest, **local_deltacat_storage_kwargs
        )

        # action
        compact_partition(
            CompactPartitionParams.of(
                {
                    "compaction_artifact_s3_bucket": TEST_S3_RCF_BUCKET_NAME,
                    "compacted_file_content_type": ContentType.PARQUET,
                    "dd_max_parallelism_ratio": 1.0,
                    "deltacat_storage": ds,
                    "deltacat_storage_kwargs": local_deltacat_storage_kwargs,
                    "destination_partition_locator": dest_partition.locator,
                    "drop_duplicates": True,
                    "hash_bucket_count": 2,
                    "last_stream_position_to_compact": source_delta.stream_position,
                    "list_deltas_kwargs": {
                        **local_deltacat_storage_kwargs,
                        **{"equivalent_table_types": []},
                    },
                    "primary_keys": ["pk"],
                    "rebase_source_partition_locator": source_delta.partition_locator,
                    "rebase_source_partition_high_watermark": source_delta.stream_position,
                    "records_per_compacted_file": 4000,
                    "s3_client_kwargs": {},
                    "source_partition_locator": source_delta.partition_locator,
                    "resource_estimation_method": ResourceEstimationMethod.PREVIOUS_INFLATION,
                }
            )
        )
